from zenaura.client.tags.node import Node, update_root_properties
from zenaura.client.tags.builder import Builder
import unittest

class TestNodeCalculatedProperties(unittest.TestCase):
    
    
    def test_level_initialization(self):
        node = Node()
        self.assertEqual(node.level, 0)  # Default level should be 0

    def test_level_propagation(self):
        root = Node("test")
        child1 = Node(children=[Node()])
        child2 = Node()
        root.children = [child1, child2]
        self.assertEqual(child1.level, 1)
        self.assertEqual(child2.level, 1)

    def test_is_leaf_initialization(self):
        node = Node()
        self.assertTrue(node.is_leaf)  # A node with no children is a leaf

        node_with_child = Node(children=[Node()])
        self.assertFalse(node_with_child.is_leaf)


    def test_key_initialization_and_assignment(self):
        root = Node()
        self.assertEqual(root.key, 0)  # Root node has key 0

        child1 = Node()
        root.children = [child1]
        self.assertEqual(child1.key, 0)  # First child under root has key 0

        child2 = Node()
        root.append_child(child2)
        self.assertEqual(child2.key, 1)  # Second child under root has key 1

    def test_key_remains_after_deletion(self):
        root = Node(children=[Node(), Node(), Node()])
        del root.children[1]  # Remove the middle child
        self.assertEqual(root.children[0].key, 0)  # Keys of remaining children are unchanged
        self.assertEqual(root.children[1].key, 2)

    def test_key_uniqueness_after_insertion(self):
        root = Node(children=[Node(), Node()])
        new_child = Node()
        root.children.insert(1, new_child)
        self.assertEqual(new_child.key, 0)  # New child gets a unique key
        self.assertEqual(root.children[2].key, 1)  # Existing child's key remains the same

    def test_level_initialization(self):
        node = Node()
        self.assertEqual(node.level, 0)  

    def test_level_propagation(self):
        root = Node()
        child1 = Node()
        child2 = Node()
        root.children = [child1, child2]
        self.assertEqual(child1.level, 1)
        self.assertEqual(child2.level, 1)
    
    def test_is_leaf_initialization(self):
        node = Node()
        self.assertTrue(node.is_leaf)  

        node_with_child = Node(children=[Node()])
        self.assertFalse(node_with_child.is_leaf)

    def test_is_text_render(self):
        node_with_text = Node(text="Hello")
        self.assertTrue(node_with_text.is_text_node)

        node_without_text = Node()
        self.assertFalse(node_without_text.is_text_node)


    def test_parent_assignment(self):
        parent = Node()
        child = Node()
        child.parent = parent
        parent.append_child(child)
        self.assertEqual(child.level, 1)

   

    def test_child_list_conversion(self):
        node = Node(children=['child1', 'child2'])
        self.assertIsInstance(node.children[0], Node)  
        self.assertEqual(node.children[0].text, 'child1')
        self.assertEqual(node.children[0].key, 0)
        self.assertEqual(node.children[1].key, 1)
        self.assertEqual(node.children[1].path, "01" )
        self.assertEqual(node.children[0].path, "00" )

    def test_is_leaf_update_on_child_change(self):
        node = Node()
        self.assertTrue(node.is_leaf)
        node.children = [Node()]
        self.assertFalse(node.is_leaf)
        node.children = []  
        self.assertTrue(node.is_leaf)


    def test_child_list_conversion(self):
        # Base Node
        root = Node(children=['child1', 'child2'])

        # Level 1 children
        child1 = root.children[0]
        child2 = root.children[1]

        # Level 2 children (added)
        child11 = Node(children=['grandchild11'])
        child12 = Node(children=['grandchild12'])
        child21 = Node(children=['grandchild21'])

        # Level 3 child (added for deeper nesting)
        great_grandchild111 = Node()

        # Attach children (to maintain the hierarchy)
        child1.append_child(child11)  # Using append
        child1.append_child(child12)
        child2.children = [child21]        # Direct assignment
        child11.append_child(great_grandchild111)
        # Assertions for base nodes (level 1)
        self.assertIsInstance(child1, Node)
        self.assertEqual(child1.text, 'child1')
        self.assertEqual(child1.key, 0)
        self.assertEqual(child1.path, "00")
        self.assertEqual(child2.path, "01")

        # Assertions for nested children (levels 2 and 3)
        self.assertEqual(child11.path, "0000")
        self.assertEqual(child12.path, "0001")
        self.assertEqual(child21.path, "0100")
        self.assertEqual(great_grandchild111.path, "000001")

    def test_builder_hierarchy(self):
        root = Builder().with_child(
            Builder().with_child(
                Builder().with_text("text").build()
            ).build()
        ).build()

        self.assertEqual(root.path, "")
        self.assertEqual(root.children[0].path, "00")
        self.assertEqual(root.children[0].children[0].path, "0010")


import sys
import unittest
from unittest.mock import patch, MagicMock
from tests.mocks.browser_mocks import MockDocument, MockWindow
from zenaura.client.tags import Node
from zenaura.client.page import Page
from zenaura.client.component import Component, Reuseable
from zenaura.client.compiler import compiler

sys.modules["pyscript"] = MagicMock()

class TestDom(unittest.TestCase):

    @patch('pyscript.document')
    @patch('pyscript.window')
    def setUp(self, document, window):  # Run before each test
        from tests.mocks.counter_mocks import Counter, BTN_STYLES, CounterState
        from zenaura.client.dom import zenaura_dom
        self.dependencies = {}  # Mock dependencies if needed
        self.btnstyles = BTN_STYLES
        self.counter = Counter(self.dependencies)
        self.counterNewInstance = Counter
        self.document = MockDocument()
        self.window = MockWindow()
        self.zenaura_dom = zenaura_dom
        self.counterState = CounterState
        # advaced dom node mock 
        self.dom_node = MagicMock()
        self.dom_node.getNodeById = MagicMock(return_value=self.dom_node)
        self.dom_node.innerHTML = ""
        
    async def test_mount(self):
        await self.zenaura_dom.mount(Page([self.counter]))
        self.assertIn(self.counter.id, self.zenaura_dom.zen_dom_table.keys())
        prev = self.zenaura_dom.zen_dom_table[self.counter.id]
        self.assertTrue(prev)
        # mount is called on root div
        exists = self.document.getElementById("root")
        self.assertTrue(exists)
	

    async def test_render(self):
        await self.zenaura_dom.mount(Page([self.counter]))
        self.counter.set_state(self.counterState(count=1))
        await self.zenaura_dom.render(self.counter)
        re_rendered = self.zenaura_dom.zen_dom_table[self.counter.id]
        # print end of line 
        # new data structure Data for data binding
        self.assertEqual(re_rendered.children[0].children[0].children[0].children[0].content, f'Counter: {self.counterState(count=1)}')

        
    async def test_mount_existing_component(self):
        
        await self.zenaura_dom.mount(Page([self.counter]))
        exists = self.zenaura_dom.zen_dom_table[self.counter.id]

        
        self.assertTrue(exists)

    async def test_render_unmounted_component(self):
        
        class K(Component):
            def render(self):
                return Node(name="div", children=[Node(name="span", children=Node(text="test"))])
        
        unmounted_counter = K()

        
        await self.zenaura_dom.render(unmounted_counter)
        key = self.zenaura_dom.zen_dom_table[self.counter.id]
        self.assertTrue(key)


        
    def test_componentDidCatchError_with_custom_error_component(self):
        # Arrange
        class CustomErrorComponent(Component):
            def __init__(self, error_message):
                super().__init__()
                self.error_message = error_message
            def render(self):
                return Node("div", children=[Node("p", children=[Node(text=self.error_message)])])

        @Reuseable
        class TestComponent(Component):
            def componentDidCatchError(self, error):
                return CustomErrorComponent(error_message=error)

        test_component = TestComponent()

        self.zenaura_dom.componentDidCatchError(test_component, "Custom error message")

        self.assertEqual(self.zenaura_dom.zen_dom_table[test_component.id].children[0].children[0].text, "Custom error message")

    def test_componentDidCatchError_with_default_error_component(self):

        class TestComponent2(Component):
            pass
        
        test = TestComponent2()
        self.zenaura_dom.componentDidCatchError(test, "Default error message")

        self.assertEqual(self.zenaura_dom.zen_dom_table[test.id].children[0].text, "Default error message")

    # Mount - testing component lifecycle methods 

    async def test_component_did_mount_without_attached_method(self):
        
        class TestComponent1(Component):
            def render(self):
                return Node("p")

        c = TestComponent1()

        await self.zenaura_dom.mount(Page([c]))

        self.assertEqual(self.zenaura_dom.zen_dom_table[c.id].name, "p")

    
    async def test_component_did_mount_with_on_seatled_method(self):

        @Reuseable
        class TestComponent3(Component):
            x = 0
    
            def attached(self, *args, **kwargs):
                self.x = 10

            def render(self):
                return Node("p")
            
        component = TestComponent3()

        self.assertEqual(component.x, 0)
        await self.zenaura_dom.mount(Page([component]))
      
        self.assertEqual(component.x, 10)


    async def test_component_did_update_with_on_settled_method(self):
        @Reuseable
        class TestComponent(Component):
            
            x  = 0
            
            def render(self):
                return Node("p")
            
            def on_settled(self, *args, **kwargs):
                self.x = 10

        component = TestComponent()
        self.assertEqual(component.x, 0)
        await self.zenaura_dom.render(component)
        self.assertEqual(component.x, 10)



import traceback
from zenaura.client.dom.error import GracefulDegenerationLifeCycleWrapper
from zenaura.client.hydrator import HydratorTasker
from .lifecycles.render import RenderLifeCycle
from zenaura.client.algorithm import DiffingAlgorithm
import asyncio


class Render(
    GracefulDegenerationLifeCycleWrapper,
    RenderLifeCycle, 
    DiffingAlgorithm,
    HydratorTasker
):
    """
    Renders components by updating the DOM based on the differences between the previous and new component trees.

    This class provides methods for:

    - **Lifecycle:** Calling lifecycle methods for components during rendering.
    - **Diffing:** Calculating the differences between the previous and new component trees.
    - **Updating:** Applying the calculated differences to the DOM.
    - **Scheduling:** Scheduling updates with the browser.
    - **Virtual DOM Update:** Updating the virtual DOM with the new component tree.
    - **Error Handling:** Handling errors gracefully and displaying an error message component.

    **Attributes:**

    - `zen_dom_table`: A dictionary that stores the virtual DOM for each component.
    - `hyd_tsk_task_queues`: A dictionary that stores task queues for each component.

    **Methods:**

    - `render(comp)`: Renders the given component by updating the DOM.
    - `on_mutation(comp)`: Calls the `on_mutation` lifecycle method for the component.
    - `on_settled(comp)`: Calls the `on_settled` lifecycle method for the component.
    - `search(prev_tree, new_tree, comp_id)`: Calculates the differences between the previous and new component trees.
    - `update(patches, comp_id)`: Applies the calculated differences to the DOM.
    - `hyd_vdom_update(comp)`: Updates the virtual DOM with the new component tree.
    - `componentDidCatchError(comp, error)`: Handles errors gracefully and displays an error message component.
    """
    async def render(self, comp ) -> None:
        """
        Renders the component by updating the DOM based on the differences between the previous and new component trees.

        This method performs the following steps:

        1. **Lifecycle:** Calls the `on_mutation` lifecycle method for the component.
        2. **Diffing:** Calculates the differences between the previous and new component trees using the `search` method.
        3. **Update:** Applies the calculated differences to the DOM using the `update` method.
        4. **Scheduling:** Schedules the updates with the browser using the `hyd_rdom_is_complete` and `hyd_tsk_dequeue_task` methods.
        5. **Virtual DOM Update:** Updates the virtual DOM with the new component tree using the `hyd_vdom_update` method.
        6. **Lifecycle:** Calls the `on_settled` lifecycle method for the component.

        **Error Handling:**

        - If an error occurs during rendering, the `componentDidCatchError` method is called with the error message.
        - This method allows components to handle errors gracefully by returning a new component to display in place of the original component.
        - If the component does not have a `componentDidCatchError` method, a default error message component is displayed.

        **Parameters:**

        - `comp`: An instance of the `Component` class.

        **Returns:**

        None
        """
        try:

            # update steps 1-3: on_mutation -> update -> on_settled
            # update 1: lifecycle method to be called before updating
            await self.on_mutation(comp)
            comp_id = comp.id            
            prev_tree = self.zen_dom_table[comp_id]
            new_tree = comp.render()

            # create task queue for component
            task_que = self.hyd_tsk_get_or_create_task_queue(comp_id)

            # run diffing algorithm
            patches = self.search(prev_tree, new_tree, comp_id)
            await self.update(patches, comp_id)
            
            # schedule update with the browser
            while not task_que.empty():
                await asyncio.sleep(0.001)
                if self.hyd_rdom_is_complete():
                    task = self.hyd_tsk_dequeue_task(comp_id)
                    await task()

                
            self.hyd_vdom_update(comp)

            # update 3  : on_settled method to be called after updating
            await self.on_settled(comp)

        except Exception as e:
            self.componentDidCatchError(comp, traceback.format_exc())


import io
import logging
import subprocess
import time
from threading import Thread, Event
import asyncio
import contextlib
from zenaura.client.page import Page 
from zenaura.client.hydrator import HydratorCompilerAdapter
from zenaura.client.app import App 
from flask import Flask
from flask_sock import Sock
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

logging.basicConfig(level=logging.INFO)
compiler_adapter = HydratorCompilerAdapter()

# create pyscript pydido template 
def template(content, meta_description=None, title=None, icon=None, pydide="https://pyscript.net/releases/2024.1.1/core.js", scripts=None):
  """
    This function generates the HTML structure of a Zenaura page.

    Args:
        content (str): The main content of the page, typically generated by compiling Zenaura components.
        meta_description (str, optional): A brief description of the page, used by search engines. Defaults to None.
        title (str, optional): The title of the page, displayed in the browser tab. Defaults to None.
        icon (str, optional): The URL of the favicon, a small icon associated with the page. Defaults to None.
        pydide (str, optional): The URL of the PyScript library, used for running Python code in the browser. Defaults to "https://pyscript.net/releases/2024.1.1/core.js".
        scripts (list, optional): An optional list of additional JavaScript scripts, CSS links to include in the page. Defaults to None.

    Returns:
        str: The complete HTML code as a string.
  """
  if scripts:
    s = io.StringIO()
    for script in scripts:
        s.write(script)
        s.write("\n")
    scripts = s.getvalue()
      
  return f"""

<html lang="en">
  <head>
    <meta charset="utf-8" />
    <link rel="icon" href="{icon}" />
    <meta name="viewport" content="width=device-width, initial-scale=1" />
    <meta name="theme-color" content="#000000" />
    <meta name="title" content="{title}" />
    <meta http-equiv="refresh"  />
    <meta
      name="description"
      content="{meta_description}"
    />
    <script type="module" src="{pydide}"></script>
    {scripts if scripts else ""}
 
	<script type="py" src="./public/main.py" config="./public/config.json"></script>

    <link  rel="stylesheet" href="./public/main.css">

    <title>{title}</title>
    
  </head>
  <body>
    <div id="root">
        {content}
    </div>
  
  </body>

</html>
"""
class ZenauraServer:
    """
    A class for server-side rendering of Zenaura applications.

    This class provides methods for:

    * Hydrating Zenaura pages for server-side rendering.
    * Hydrating Zenaura apps for server-side rendering.
    * Generating the HTML structure of a Zenaura page.
    """

    @staticmethod
    def hydrate_page(page: Page, title="zenaura", meta_description="this app created with zenaura", icon="./public/favicon.ico", pydide="https://pyscript.net/releases/2024.1.1/core.js") -> str:
        """
        Hydrates a Zenaura page for server-side rendering.

        This method compiles the page's components using the HydratorCompilerAdapter and generates the HTML structure of the page.

        Args:
            page (Page): The Zenaura page to be hydrated.
            title (str, optional): The title of the page. Defaults to "zenaura".
            meta_description (str, optional): The meta description of the page. Defaults to "this app created with zenaura".
            icon (str, optional): The URL of the favicon. Defaults to "./public/favicon.ico".
            pydide (str, optional): The URL of the PyScript library. Defaults to "https://pyscript.net/releases/2024.1.1/core.js".

        Returns:
            str: The HTML structure of the hydrated page.
        """

        return template(compiler_adapter.hyd_comp_compile_page(page), meta_description, title, icon, pydide)

    @staticmethod
    def hydrate_app(app: App, title="zenaura", meta_description="this app created with zenaura", icon="./public/favicon.ico", pydide="https://pyscript.net/releases/2024.1.1/core.js", scripts=None) -> None:
        """
        Hydrates a Zenaura app for server-side rendering.

        This method renders all pages in the app, sets the page with path "/" to visible, and the rest to hidden. It then compiles the index.html file for server-side rendering.

        Args:
            app (App): The Zenaura app to be hydrated.
            title (str, optional): The title of the page. Defaults to "zenaura".
            meta_description (str, optional): The meta description of the page. Defaults to "this app created with zenaura".
            icon (str, optional): The URL of the favicon. Defaults to "./public/favicon.ico".
            pydide (str, optional): The URL of the PyScript library. Defaults to "https://pyscript.net/releases/2024.1.1/core.js".
            scripts (list, optional): An optional list of additional JavaScript scripts and CSS links to include in the page. Defaults to None.
        """

        pages = io.StringIO()

        # Render pages
        for path, route in app.routes.items():
            page, _, _, ssr = route
            if ssr:  # Ignore SSR pages
                continue
            if path == "/":  # Set / route to visible
                page_div = lambda comps: f'<div data-zenaura="{page.id}">{comps}</div>'
                pages.write(page_div(compiler_adapter.hyd_comp_compile_page(page)))
                continue
            # Pages other than / are set to hidden
            page_div = lambda comps: f'<div hidden data-zenaura="{page.id}">{comps}</div>'
            pages.write(page_div(compiler_adapter.hyd_comp_compile_page(page)))

        pages = pages.getvalue()

        # Overwrite in public dir
        with open("./public/index.html", "w") as file:
            file.write(template(pages, meta_description, title, icon, pydide, scripts))


class PausingObserver(Observer):
    def dispatch_events(self, *args, **kwargs):
        if not getattr(self, '_is_paused', False):
            super(PausingObserver, self).dispatch_events(*args, **kwargs)

    def pause(self):
        self._is_paused = True

    def resume(self):
        time.sleep(self.timeout)  # allow interim events to be queued
        self.event_queue.queue.clear()
        self._is_paused = False

    @contextlib.contextmanager
    def ignore_events(self):
        self.pause()
        yield
        self.resume()

class DevServer:
    """
    A class for running a development server for Zenaura applications.

    This class provides methods for:

    * Starting a Flask server with WebSocket support.
    * Sending refresh signals to connected clients when changes are detected.
    * Hydrating the application and notifying clients when changes are made.
    * Running a file system observer to detect changes in the application files.
    """

    def __init__(self, debug=True, port=5000):
        """
        Initializes the DevServer class.

        Args:
            debug (bool, optional): Whether to run the server in debug mode. Defaults to True.
            port (int, optional): The port on which to run the server. Defaults to 5000.
        """

        self.debug = debug
        self.port = port
        self.app = Flask(__name__, static_folder="public", template_folder="public")
        self.sock = Sock()
        self.ws_client_list = []
        self.shutdown_event = Event()
        self.observer = PausingObserver()
        self.sock.init_app(self.app)
        self.loop = asyncio.new_event_loop()

        self.setup_websocket()

    def setup_websocket(self):
        """
        Sets up the WebSocket route for sending refresh signals to clients.
        """

        @self.sock.route("/refresh")
        def refresh(ws):
            """
            WebSocket handler for sending refresh signals to clients.

            Args:
                ws (WebSocket): The WebSocket connection.
            """

            self.ws_client_list.append(ws)
            while not self.shutdown_event.is_set():
                try:
                    ws.receive()
                    ws.sleep(1)
                    ws.send("refresh")
                except Exception as e:
                    print(f"Error in WebSocket connection: {e}")
                    break

    def send_refresh_signal(self):
        """
        Sends a refresh signal to all connected clients.
        """

        logging.info("Sending refresh signal...")
        clients = self.ws_client_list.copy()
        for client in clients:
            try:
                client.send("refresh")
            except Exception as e:
                print(f"Error sending refresh: {e}")
                self.ws_client_list.remove(client)

    def get_change_handler(self):
        """
        Returns a ChangeHandler class that handles file system events.

        Returns:
            ChangeHandler: A class that handles file system events.
        """

        DEVSERVER = self

        class ChangeHandler(FileSystemEventHandler):
            """
            A class that handles file system events.

            This class is used to detect changes in the application files and trigger a refresh of the browser.
            """

            def __init__(self, server):
                """
                Initializes the ChangeHandler class.

                Args:
                    server (DevServer): The DevServer instance.
                """

                super().__init__()
                self.server = server

            def on_any_event(self, event):
                """
                Handles file system events.

                Args:
                    event (FileSystemEvent): The file system event.
                """

                try:
                    logging.info(f"File {event.src_path} has changed.")
                    logging.info("Changes are live...")
                    DEVSERVER.hydrate_and_notify()
                    logging.info("Reloading browser...")
                    DEVSERVER.send_refresh_signal()
                    logging.info("Browser reloaded.")
                except Exception as e:
                    logging.info(f"Error in ChangeHandler: {e}")

        return ChangeHandler

    def start_server(self):
        """
        Starts the Flask server.
        """

        try:
            self.app.run(debug=self.debug, port=self.port, use_reloader=False)
        except Exception as e:
            logging.info(f"Error starting server: {e}")

    def hydrate_and_notify(self):
        """
        Hydrates the application and notifies clients of changes.
        """

        try:
            self.observer.pause()
            logging.info("Hydrating...")
            logging.info("Pausing the observer...")
            process = subprocess.Popen("python build.py", shell=True)
            process.communicate()
            logging.info("Hydrated done...")

        finally:
            logging.info("Running the observer...")
            self.observer.resume()

    def run(self):
        """
        Runs the development server.

        This method starts the Flask server, file system observer, and WebSocket server.
        """

        path = 'public'
        ChangeHandler = self.get_change_handler()
        event_handler = ChangeHandler(self)
        self.observer.schedule(event_handler, path, recursive=True)
        self.observer.start()

        server_thread = Thread(target=self.start_server, daemon=True)
        server_thread.start()

        try:
            while not self.shutdown_event.is_set():
                time.sleep(0.1)  # Shorter sleep interval
        except KeyboardInterrupt:
            logging.info("KeyboardInterrupt received, stopping...")
        finally:
            # Faster Shutdown of Observer
            self.observer.event_queue.queue.clear()
            self.observer.stop()  
            self.observer.join()  


class Attribute:
    """
    Represents a key-value pair used for tagging entities.

    Attributes:
        key (str): The key of the attribute.
        value (str): The value of the attribute.

    Methods:
        to_dict(): Converts the attribute to a dictionary representation.
    """

    def __init__(self, key, value):
        """
        Initializes an Attribute object with the given key and value.

        Args:
            key (str): The key of the attribute.
            value (str): The value of the attribute.
        """
        self.key = key
        self.value = value

    def to_dict(self) -> dict:
        """
        Returns a dictionary representation of the attribute.

        Returns:
            dict: A dictionary containing the key and value of the attribute.
        """
        return {
            "key": self.key,
            "value": self.value
        }



class RenderLifeCycle:
    """
    This class provides lifecycle methods for components that are rendered to the DOM.

    It allows components to perform actions before and after they are updated and re-rendered in the DOM.

    Attributes:
        None
    """

    async def on_mutation(self, comp) -> None:
        """
        This method is called after the component is updated in the DOM and re-rendered.

        It allows the component to perform any necessary actions before the update is applied, such as:

        - Updating state based on new props
        - Setting up event listeners
        - Making API calls
        - Performing animations

        Args:
            comp: An instance of the Component class.

        Returns:
            None
        """

        # Perform operations before updating
        if hasattr(comp, 'on_mutation'):
            await comp.on_mutation()

    async def on_settled(self, comp) -> None:
        """
        This method is called after the component is updated in the DOM and re-rendered.

        It allows the component to perform any necessary actions after the update is applied, such as:

        - Focusing on an input element
        - Scrolling to a specific position
        - Triggering custom events

        Args:
            comp: An instance of the Component class.

        Returns:
            None
        """

        # Perform operations after updating
        if hasattr(comp, 'on_settled'):
            await comp.on_settled()



import asyncio
from dataclasses import dataclass
from typing import List
from zenaura.client.dom import zenaura_dom
from zenaura.client.tags import Node
from zenaura.client.component import Component, Reuseable
from zenaura.client.page import Page
from zenaura.client.mocks import MockWindow, MockDocument
from zenaura.client.hydrator import HydratorRealDomAdapter

rdom_hyd = HydratorRealDomAdapter() 

event_loop = asyncio.get_event_loop()

# this is really nothing just to be able to mock 
try :
    from pyscript import document, window
except ImportError:
    document = MockDocument() 
    window = MockWindow() 
    
from typing import Optional, Tuple, Callable, Dict, Any 
from collections import defaultdict

class HistoryNode:
    def __init__(self, page=None):
        self.page = page
        self.prev = None
        self.next = None 

class PageHistory:
    def __init__(self):
        self.current = HistoryNode()

    def visit(self, page : Page) -> None:
        new_page  =HistoryNode(page)
        new_page.prev = self.current
        self.current.next = new_page
        self.current = new_page

    def back(self) -> None:
        if self.current.prev:
            self.current = self.current.prev
        return self.current.page
    
    def forward(self) -> None:
        if self.current.next:
            self.current = self.current.next
        return self.current.page

@Reuseable
class NotFound(Component):
    """
    Represents a page for displaying a "page not found" message.

    Methods
    -------
    render()
        Returns a Node representing the "page not found" message.
    """
    def render(self):
        em = Node("div")
        em.append_child(Node(name="text", children=["page not found"]))
        return em
    
notFound = NotFound()


class Route:
    """
        Represents a route configuration for the App.
    """
    def __init__(self, title, path, page, middleware=None, ssr=False):
        """
        Initializes a Route with the specified title, path, and pageonent.
        Attributes
        ----------
        title : str
            The title of the route.
        path : str
            The path of the route.
        page : Page
            A page of pageonents.
        handler : Optional[Callable]
            Optional route-specific logic handler.
        middleware : Optional[Callable]
            Optional route-specific logic handler.
        ssr : bool
            Whether the route is server-side rendered.
        """
        self.title = title
        self.path = path
        self.page = page
        self.ssr=ssr

        if not isinstance(page, Page):
            raise TypeError("Only a Page can be mounted on a route")
        self.middleware: Optional[Callable] = middleware  # For optional route-specific logic


# router 
class App:
    """
    Represents a router for managing routes and navigation.

    This class provides methods for adding routes, navigating between pages, and handling the current location.

    Attributes:
        routes (dict): A dictionary mapping paths to their associated pages and titles.
        paths (list): A list of paths registered in the router.
        history (PageHistory): An object that manages the history of visited pages.

    Methods:
        __init__()
            Initializes the App with empty routes and paths, and sets up the initial route handling.
        navigate(path)
            Navigates to the specified path by mounting the associated page and updating the document title and browser history.
        handle_location()
            Handles the current location by mounting the associated page and updating the document title.
        add_route(route)
            Adds a route to the router's configuration.
        back()
            Navigates back to the previous Page in the history stack.
        forward()
            Navigates forward to the next Page in the history stack.
        get_current_route()
            Get the page and title of the current route, or None if not found.
    """
    def __init__(self):
        """
            Initializes the App with empty routes and paths, and sets up the initial route handling.
        """
        # key -> path , value -> [page, document.title]
        self.routes = defaultdict(str)
        self.paths = []
        self.history = PageHistory()

    async def not_found(self):
        document.title = "Page Not Found"
        page = Page([notFound])
        await zenaura_dom.mount(page)
        self.history.visit(page)

    async def navigate(self, path) -> None:
        """
        Navigates to the specified path by mounting the associated pageonent and updating the document title and browser history.

        Parameters
        ----------
        path : str
            The path to navigate to.
        """
        matched_route, params = self._match_route(path)
        if not matched_route:
            await self.not_found()
            return
        
        [page, title, middleware, ssr] = self.routes[path]

        if callable(middleware):
            await middleware()

        if ssr:  # Ignore mount step for server side rendering pages.
            await zenaura_dom.mount(page)
            self.history.visit(page)
            document.title = title
            return

        if not self.history.current.page:  # self.history.current is initially None
            pass
        else:
            self.hyd_rdom_toggle_pages_visibilty(self.history.current.page, page)
        
        window.history.pushState(path, title, path)  # Update browser history
        await zenaura_dom.mount(page)
        self.history.visit(page)
        document.title = title

    def hyd_rdom_toggle_pages_visibilty(self, previous_page: Page, current_page: Page):
        p_page = document.querySelector(f'[data-zenaura="{previous_page.id}"]')
        if p_page:
            p_page.hidden = True  # Hide the previous page
        curr_page = document.querySelector(f'[data-zenaura="{current_page.id}"]')
        if curr_page:
            curr_page.hidden = False  # Show the current page

    async def handle_location(self) -> None:
        """
        Handles the current location by mounting the associated page and updating the document title.
        """
        path = window.location.pathname
        matched_route, params = self._match_route(path)
        if not matched_route:
            await self.not_found()
            return
        [page, title, middleware, ssr] = self.routes[path]
        window.history.pushState(path, title, path)  # Update browser history

        if callable(middleware):
            await middleware()
        if ssr:  # Ignore mount step for server side rendering pages.
            await zenaura_dom.mount(page)
            self.history.visit(page)
            document.title = title
            return
        if not self.history.current.page:  # self.history.current is initially None
           pass
        else:
            self.hyd_rdom_toggle_pages_visibilty(self.history.current.page, page)
        self.history.visit(page)
        await zenaura_dom.mount(page)  # Trigger attached lifecycle for each component within the page.


    def add_route(self, route : Route) -> None:
        """
        Adds a route to the router's configuration.

        Parameters
        ----------
        route : Route
            The route to be added to the router's configuration.
        """
        self.routes[route.path] = [route.page, route.title, route.middleware, route.ssr]
        self.paths.append(route.path)
    
    async def back(self) -> None:
        """
        Navigates back to the previous Page in the history stack.
        """
        previous_page = self.history.current.page
        curr_page = self.history.back()
        if not curr_page.ssr: # ignore mount step for server side rendering pages.
            rdom_hyd.hyd_rdom_toggle_pages_visibilty(previous_page, curr_page)
            await zenaura_dom.mount(curr_page)  # trigger attached lifecycle for each component within the page.
        else: # trigger attached lifecycle method for the component.
            await zenaura_dom.mount(curr_page)
    
    async def forward(self) -> None:
        """
        Navigates forward to the next Page in the history stack.
        """
        previous_page = self.history.current.page
        curr_page = self.history.forward()
        if not curr_page.ssr: # ignore mount step for server side rendering pages.
            rdom_hyd.hyd_rdom_toggle_pages_visibilty(previous_page, curr_page)
            await zenaura_dom.mount(curr_page)  # trigger attached lifecycle for each component within the page.
        else: # trigger attached lifecycle method for the component.
            await zenaura_dom.mount(curr_page)

    def get_current_route(self) -> Optional[Tuple[Page, str]]:
        """
        Get the page and title of the current route, or None if not found.
        """
        path = window.location.pathname
        matched_route, info = self._match_route(path)
        return matched_route, info
    
    # TODO still needs a lot of work
    def _match_route(self, path: str) -> Tuple[Optional[Tuple[Page, str, Dict[str, Any]]], Dict[str, str]]:
        """
        Matches the given path to a registered route and extracts parameters.

        Args:
            path (str): The path to match.

        Returns:
            Tuple[Optional[Tuple[Page, str, Dict[str, Any]]], Dict[str, str]]:
                A tuple containing the matched route information (if any) and extracted parameters.
        """
        for route_path, (page, title, middleware, ssr) in self.routes.items():
            if "*" in route_path:  # Wildcard route
                route_parts = route_path.split("*")
                if path.startswith(route_parts[0]):
                    params = path[len(route_parts[0]):]
                    query = defaultdict(str)
                    if "?" in params: # get all queries
                        newq = "".join(params).split("?")
                        params = newq[0]
                        newq = newq[1]
                        for q in newq.split("&"):
                            k,v = q.split("=")
                            query[k] = v
                    if "/" in params: # get all params
                        params = params.split("/")
                   
                    params = {"wildcard": {"params" : params, "query" : query} }

                    return (page, title, middleware, ssr), params
            elif ":" in route_path:  # Parameterized route
                route_parts = route_path.split("/")
                path_parts = path.split("/")
                if len(route_parts) == len(path_parts):
                    params = {}
                    for i, part in enumerate(route_parts):
                        if part.startswith(":"):
                            param_name = part[1:]
                            params[param_name] = path_parts[i]
                    return (page, title, middleware, ssr), params
            elif route_path == path:  # Exact match
                return (page, title, middleware, ssr), {}
        return None, {}  # No match found
        
    #TODO transition effects 


class Subject:
    """
    Create subjects for components to communicate on.

    A subject is an entity that maintains a state and notifies its observers when the state changes.
    Observers can be any object that implements the `Observer` interface.

    **Key Features:**

    - **State Management:** Subjects maintain their own state and provide access to it through the `state` property.
    - **Observer Management:** Subjects manage a list of attached observers and notify them when the state changes.
    - **Loose Coupling:** Subjects are not directly dependent on the implementation of their observers, 
      promoting flexibility and maintainability.

    **Usage:**

    1. Create a subject instance.
    2. Attach observers to the subject using the `attach` method.
    3. Modify the subject's state using the `state` property.
    4. The subject will automatically notify all attached observers when the state changes.

    **Example:**

    ```python
    class ConcreteSubject(Subject):
        def __init__(self, initial_state):
            super().__init__()
            self._state = initial_state

        def do_something(self):
            # Modify the subject's state
            self.state = {"key": "value"}

    # Create a subject and an observer
    subject = ConcreteSubject({"initial": "value"})
    observer = Observer()

    # Attach the observer to the subject
    subject.attach(observer)

    # Modify the subject's state
    subject.do_something()
    ```

    **Additional Notes:**

    - The `state` property can be any type of data, depending on the subject's implementation.
    - Observers can choose to ignore notifications if they are not interested in the updated state.
    - The subject can manage multiple observers and notify them individually or in groups.
    """

    def __init__(self):
        """
        Initialize a new Subject.

        Parameters:
        None

        Returns:
        None
        """
        self._observers = set()
        self._state = {}

    def attach(self, observer):
        """
        Attach an observer to the subject.

        Parameters:
        observer (Observer): The observer to attach.

        Returns:
        None
        """
        self._observers.add(observer)

    def detach(self, observer):
        """
        Detach an observer from the subject.

        Parameters:
        observer (Observer): The observer to detach.

        Returns:
        None
        """
        self._observers.discard(observer)

    def notify(self):
        """
        Notify all attached observers.

        Parameters:
        None

        Returns:
        None
        """
        for observer in self._observers:
            observer.update(self._state)

    @property
    def state(self):
        """
        Get the state of the subject.

        Parameters:
        None

        Returns:
        dict: The state of the subject.
        """
        return self._state

    @state.setter
    def state(self, new_value):
        """
        Set the state of the subject and notify observers.

        Parameters:
        new_value (dict): The new state of the subject.

        Returns:
        None
        """
        self._state = new_value
        self.notify()



#!/usr/bin/env python3
#!/usr/bin/env python3
import itertools
import hashlib
from abc import abstractmethod
from collections import defaultdict

_is_reuseable = defaultdict(lambda: False)


class UUIDManager:
    @staticmethod
    def generate_uuid(cls_name, count):
        # Combine class name and count to create a unique string
        unique_string = f"{cls_name}{count}"
        # Hash the unique string to produce a 32-bit UUID
        uuid_hash = hashlib.md5(unique_string.encode()).hexdigest()[:8]
        return uuid_hash

def Reuseable(cls):
    """
    Decorator to mark a component as reusable.

    Reusable components can be instantiated multiple times and will maintain their own state.

    Args:
        cls (type): The component class to be decorated.

    Returns:
        type: The decorated component class.
    """

    original_init = cls.__init__

    def new_init(self, *args, **kwargs):
        original_init(self, *args, **kwargs)
        cls.count = next(cls._component_count)
        self.id = UUIDManager.generate_uuid(cls.__name__, self.count)
        _is_reuseable[cls.__name__] = True

    cls.__init__ = new_init
    return cls

class SelectiveSingleton(type):
    """
    Metaclass to ensure only one instance of a component exists at a time.

    This metaclass is used for components that should only have one instance, even if they are instantiated multiple times.
    """
    _instances = defaultdict(str)

    def __call__(cls, *args, **kwargs):
        # Check if the class has the _is_reuseable attribute set to True
        if not _is_reuseable[cls.__name__]:
            return super().__call__(*args, **kwargs)

        reuseableId = f"{cls.__name__ + str(cls.count)}"   
        if cls not in cls._instances and reuseableId not in cls._instances and _is_reuseable[cls.__name__]:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        else:
            instance = super().__call__(*args, **kwargs)
            cls._instances[reuseableId] = instance
        return cls._instances[cls]
    
class Component(metaclass=SelectiveSingleton):
    """
    Base class for all Zenaura components.

    Components are the building blocks of Zenaura applications. They represent reusable units of functionality that can be composed to create complex user interfaces.

    Attributes:
        id (str): A unique identifier for the component.
        state (dict): The state of the component.
        _state (dict): The internal state of the component.
        _track_instances (dict): A dictionary tracking the number of instances created for each component class.
        _component_count (itertools.count): An iterator that generates unique counts for each component instance.

    Methods:
        __init_subclass__(cls, **kwargs):
            Initializes the subclass and sets the initial count for the component class.
        __init__(self):
            Initializes the component instance and sets the unique identifier.
        get_state(self):
            Returns the state of the component.
        set_state(self, state):
            Sets the state of the component.
        render(self):
            Abstract method that must be implemented by subclasses to define the behavior of the component.
    """

    _state = defaultdict(str)
    _track_instances = defaultdict(int)
    _component_count = itertools.count(0)

    def __init_subclass__(cls, **kwargs):
        """
        Initializes the subclass and sets the initial count for the component class.
        """
        cls.count = next(cls._component_count)
        cls.id = UUIDManager.generate_uuid(cls.__name__, cls.count)
        super().__init_subclass__(**kwargs)
        
    def __init__(self):
        """
        Initializes the component instance and sets the unique identifier.
        """
        cls = self.__class__
        Component._track_instances[cls.__name__] += 1
        if Component._track_instances[cls.__name__] > 1 and not _is_reuseable[cls.__name__]:
            raise TypeError(
"""
    Zenaura class component are limted by design. \n
    Decorate component with @Reuseable to implicitly  \n
    state the component is meant to be reused:  \n
    example :  \n
        class ThisIsLimited(Component):  \n
            pass
        c1 = ThisIsLimited() // no error  \n
        c2 = ThisIsLimited() // throws error  \n

        @Reuseable  \n
        class ThisIsReuseable(Component):  \n
        c1 = ThisIsReuseable() // no error  \n
        c2 = ThisIsReuseable() // no error  \n
"""
            )
        
        
        # print(Component._track_instances[cls.__name__], cls.__name__, is_decorated_with_reuseable(cls))
       
    @property
    def state(self):
        """
        Get the state of the component.

        Returns:
        dict: The state of the component.
        """

        return self.get_state()

    @state.setter
    def state(self, value):
        """
        Set the state of the component.

        Args:
        value (dict): The new state of the component.

        Returns:
        None
        """

        self.set_state(value)

    def get_state(self):
        """
        Get the state of the component.

        Returns:
        dict: The state of the component.
        """

        return self._state

    def set_state(self, state):
        """
        Set the state of the component.

        Args:
        state (dict): The new state of the component.

        Returns:
        None
        """

        self._state = state  # Update the internal state

    @abstractmethod
    def render():
        """
        Abstract method to be implemented by subclasses.

        This method should be implemented by subclasses to define the behavior of the component.

        Returns:
        None
        """


from .searcher import Searcher
from .updater import Updater

class DiffingAlgorithm(
    Searcher,
    Updater
):
    """
        The diffing algorithm in Zenaura virtual DOM.

        This class is responsible for comparing the old and new virtual DOM trees and generating a list of changes that need to be applied to the real DOM.

        The diffing algorithm is implemented in two steps:

        1. **Search:** The `Searcher` class traverses the old and new virtual DOM trees and identifies the nodes that have changed.
        2. **Update:** The `Updater` class applies the changes to the real DOM.

        The diffing algorithm is designed to be efficient and performant. It uses a number of techniques to minimize the number of changes that need to be applied to the real DOM.

        **Attributes:**

        * `Searcher`: An instance of the `Searcher` class.
        * `Updater`: An instance of the `Updater` class.

        **Methods:**

        * `diff(old_tree, new_tree)`: Compares the old and new virtual DOM trees and generates a list of changes that need to be applied to the real DOM.

        **Example:**

        ```python
        # Create an instance of the DiffingAlgorithm class.
        diffing_algorithm = DiffingAlgorithm()

        # Compare the old and new virtual DOM trees.
        changes = diffing_algorithm.diff(old_tree, new_tree)

        # Apply the changes to the real DOM.
        diffing_algorithm.update(changes)
        ```
    """


from zenaura.client.tags import Node
from typing import List
from zenaura.client.hydrator.compiler_adapter import HydratorCompilerAdapter
from itertools import zip_longest
from .operations import *

class Searcher(
    HydratorCompilerAdapter
):
    """
    This class implements the searching step of the Zenaura virtual DOM algorithm.

    It takes two virtual DOM trees, the previous and the new one, and identifies the differences between them.
    These differences are then used to update the real DOM efficiently.

    Attributes:
        None
    """

    def updater_context_builder(self, name: str, context: dict) -> dict:
        """
        Builds the context for the updater.

        This method takes the name of the operation and its context and returns a dictionary that will be used by the updater.

        Args:
            name: The name of the operation.
            context: The context of the operation.

        Returns:
            A dictionary containing the name and context of the operation.
        """

        return {
            "name": name,
            "context": context
        }

    def patches_builder(self, prev_child_node, new_child_node, id, child_id):
        """
        Builds the patches for the updater.

        This method takes the previous and new child nodes, the component ID, and the child ID, and returns a list of patches that will be used by the updater.

        Args:
            prev_child_node: The previous child node.
            new_child_node: The new child node.
            id: The ID of the component.
            child_id: The ID of the child.

        Returns:
            A list of patches containing the operation name, the new child node, the child ID, and the context for the updater.
        """

        return [
            self.hyd_comp_get_keyed_uuid(
                id=id,
                child_id=child_id
            ),
            new_child_node,
            child_id,
            self.updater_context_builder(
                name=REMOVE_NODE,
                context={"children": prev_child_node}
            )
        ]

    def search(self, prevNode: Node, newNode: Node, id: str) -> List[List[any]]:
        """
        Searches for the differences between the previous and new virtual DOM trees.

        This method takes the previous and new virtual DOM trees and the component ID, and returns a list of lists of differences. Each difference is represented as a list containing the operation name, the new child node, the path of the child, and the context for the updater.

        Args:
            prevNode: The previous virtual DOM tree.
            newNode: The new virtual DOM tree.
            id: The ID of the component.

        Returns:
            A list of lists of differences.
        """

        # Error handling:
        if not prevNode and not newNode:
            return []
        if not isinstance(prevNode, Node) or not isinstance(newNode, Node):
            return []

        differences = []

        def helper(prev_child_node: Node, new_child_node: Node, id, prev_child_path: str, new_child_path) -> None:
            """
            Helper method for the search method.

            This method recursively compares the previous and new child nodes and their attributes and children, and adds the differences to the `differences` list.

            Args:
                prev_child_node: The previous child node.
                new_child_node: The new child node.
                id: The ID of the component.
                prev_child_path: The path of the previous child node.
                new_child_path: The path of the new child node.
            """

            nonlocal differences

            # Added node
            if not prev_child_node and new_child_node:
                differences.append([
                    self.hyd_comp_get_keyed_uuid(
                        id=id,
                        key=prev_child_path
                    ),
                    new_child_node,
                    prev_child_path,
                    self.updater_context_builder(
                        name=ADD_NODE,
                        context={"children": new_child_node}
                    )
                ])
                return

            # Removed node
            if prev_child_node and not new_child_node:
                differences.append([
                    self.hyd_comp_get_keyed_uuid(
                        id=id,
                        key=prev_child_path
                    ),
                    prev_child_node,
                    prev_child_path,
                    self.updater_context_builder(
                        name=REMOVE_NODE,
                        context={"children": prev_child_node}
                    )
                ])
                return

            # Changed child by name
            if prev_child_node.name != new_child_node.name:
                differences.append([
                    self.hyd_comp_get_keyed_uuid(
                        id=id,
                        key=prev_child_path
                    ),
                    prev_child_node,
                    prev_child_path,
                    self.updater_context_builder(
                        name=REMOVE_NODE,
                        context={"children": prev_child_node}
                    )
                ])
                differences.append([
                    self.hyd_comp_get_keyed_uuid(
                        id=id,
                        key=new_child_path
                    ),
                    new_child_node,
                    new_child_path,
                    self.updater_context_builder(
                        name=ADD_NODE,
                        context={"children": new_child_node}
                    )
                ])
                return

            # Compare attributes
            for [prev_attr, new_attr] in zip_longest(prev_child_node.attributes, new_child_node.attributes):

                # Removed attribute
                if prev_attr and not new_attr:
                    differences.append([
                        self.hyd_comp_get_keyed_uuid(
                            id=id,
                            key=prev_child_path
                        ),
                        new_child_node,
                        prev_child_path,
                        self.updater_context_builder(
                            name=REMOVE_ATTRIBUTE,
                            context={"attr_name": prev_attr.key}
                        )
                    ])
                    continue

                # Added attribute
                if not prev_attr and new_attr:
                    differences.append([
                        self.hyd_comp_get_keyed_uuid(
                            id=id,
                            key=new_child_path
                        ),
                        new_child_node,
                        new_child_path,
                        self.updater_context_builder(
                            name=ADD_ATTRIBUTE,
                            context={"attr_name": new_attr.key, "attr_value": new_attr.value}
                        )
                    ])
                    continue

                # Replaced value attribute
                if prev_attr and new_attr:
                    if prev_attr.value != new_attr.value:
                        differences.append([
                            self.hyd_comp_get_keyed_uuid(
                                id=id,
                                key=prev_child_path
                            ),
                            new_child_node,
                            prev_child_path,
                            self.updater_context_builder(
                                name=ADD_ATTRIBUTE,
                                context={"attr_name": new_attr.key, "attr_value": new_attr.value}
                            )
                        ])

            # Compare children
            for idx, (prev_child, new_child) in enumerate(zip_longest(prev_child_node.children, new_child_node.children)):
                # Leaf text nodes
                if prev_child and new_child:
                    if prev_child.is_text_node and new_child.is_text_node and (prev_child.text != new_child.text):
                        differences.append([
                            self.hyd_comp_get_keyed_uuid(
                                id=id,
                                key=prev_child_path
                            ),
                            new_child_node,
                            prev_child_path,
                            self.updater_context_builder(
                                name=NODE_INNER_TEXT,
                                context={"text": new_child.text}
                            )
                        ])
                        continue

                new_child_path = new_child.path if isinstance(new_child, Node) else ""
                prev_child_path = prev_child.path if isinstance(prev_child, Node) else ""
                helper(prev_child, new_child, id, prev_child_path=prev_child_path, new_child_path=new_child_path)

        helper(prevNode, newNode, id, 0, 0)
        return differences


import html
import bleach

from zenaura.client.config import (
    allowed_tags,
    allowed_attributes,
)

class CompilerSanitizer:
    """
    This class provides methods for sanitizing user input to prevent various injection attacks.

    It uses the `bleach` library to remove potentially harmful HTML tags and attributes, 
    and the `html` library to escape special characters.

    Attributes:
        None
    """

    def sanitize(
        self, 
        user_input: str, 
        allowed_tags: list = allowed_tags, 
        allowed_attributes: dict = allowed_attributes
    ) -> str:
        """
        Sanitizes user input to prevent various injection attacks.

        This method takes the raw user input as a string and returns a sanitized version of the input.

        Args:
            user_input (str): The raw user input to sanitize.
            allowed_tags (list, optional): A list of allowed HTML tags (e.g., ['p', 'br', 'strong']). 
                Defaults to `allowed_tags` from the `zenaura.client.config` module.
            allowed_attributes (dict, optional): A dictionary mapping allowed tags to their allowed 
                attributes (e.g., {'img': ['src', 'alt']}). Defaults to `allowed_attributes` from the 
                `zenaura.client.config` module.

        Returns:
            str: The sanitized input.
        """

        # Convert user input to string for safety
        user_input = str(user_input)

        # Escape all HTML special characters initially
        safe_html = html.escape(user_input)

        # Use bleach to remove potentially harmful HTML tags and attributes
        safe_html = bleach.clean(
            safe_html, tags=allowed_tags, attributes=allowed_attributes
        )

        return safe_html


from zenaura.client.hydrator import HydratorVirtualDomAdapter

class MountLifeCycles(
    HydratorVirtualDomAdapter
):
    """
    This class provides lifecycle methods for components that are mounted to the DOM.

    It inherits from the `HydratorVirtualDomAdapter` class, which provides methods for interacting with the virtual DOM.
    """

    async def attached(self, comp) -> None:
        """
        This method is called after the component is mounted to the DOM.

        It allows the component to perform any necessary actions after it has been added to the DOM, such as:

        - Initializing state
        - Setting up event listeners
        - Making API calls
        - Performing animations

        Args:
            comp: An instance of the Component class.

        Returns:
            None
        """

        if hasattr(comp, 'attached'):
            await comp.attached()



import unittest
import io
from zenaura.client.page import Page 
from .mocks.counter_mocks import Counter
from zenaura.client.hydrator import HydratorCompilerAdapter
from zenaura.server.server import ZenauraServer, template
from zenaura.client.app import App 
from unittest.mock import MagicMock, patch


class TestZenauraServer(unittest.TestCase):

    def setUp(self):
        self.page = Page([Counter([])])
        self.page.id = 'test-page-id'
        self.page_content = '<div>Test Page Content</div>'
        self.compiler_adapter = HydratorCompilerAdapter()
        self.maxDiff = None
        
        global compiler_adapter
        compiler_adapter = self.compiler_adapter  # Mocking the global instance
        
        self.template_content = template

    def test_hydrate_page(self):
        result = ZenauraServer.hydrate_page(self.page)
        self.assertEqual(result, self.template_content(self.compiler_adapter.hyd_comp_compile_page(self.page),  title="zenaura", meta_description="this app created with zenaura", icon="./public/favicon.ico", pydide="https://pyscript.net/releases/2024.1.1/core.js"))
    

if __name__ == '__main__':
    unittest.main()


from zenaura.client.dom import zenaura_dom
import functools

def mutator(func):
    """
    Decorator to automatically render the DOM after the decorated function executes.

    This decorator is useful for functions that modify the state of the application,
    triggering a re-rendering of the DOM to reflect the changes.

    Args:
        func: The function to be decorated.

    Returns:
        wrapper_func: A wrapper function that executes the original function and then renders the DOM.

    Raises:
        TypeError: If `func` is not a callable.
    """

    if not callable(func):
        raise TypeError("mutator can only decorate functions")

    @functools.wraps(func)
    async def wrapper_func(self, *args, **kwargs):
        """
        Wrapper function that executes the original function and then renders the DOM.

        Args:
            self: The instance of the class.
            *args: Positional arguments for the original function.
            **kwargs: Keyword arguments for the original function.

        Returns:
            None
        """

        await func(self, *args, **kwargs)
        await zenaura_dom.render(self)

    return wrapper_func



import sys
import unittest
from unittest.mock import patch, MagicMock
from zenaura.client.tags import Attribute, Node
from tests.mocks.browser_mocks import MockDocument, MockWindow

sys.modules["pyscript"] = MagicMock()

class TestComponent(unittest.TestCase):

	@patch('pyscript.document')
	@patch('pyscript.window')
	def setUp(self, document, window):  # Run before each test
		from tests.mocks.counter_mocks import Counter, BTN_STYLES, CounterState
		from zenaura.client.dom import zenaura_dom
		self.dependencies = {}  # Mock dependencies if needed
		self.btnstyles = BTN_STYLES
		self.counter = Counter(self.dependencies)
		document = MockDocument()
		window = MockWindow()
		self.zenaura_dom = zenaura_dom
		self.counterState = CounterState

	def test_initial_state(self):
		state = self.counter.get_state()
		self.assertEqual(state.count, 0)  # Ensure initial count is 0
	

	async def test_increment(self):
		await self.counter.increment()
		state = self.counter.get_state()
		self.assertEqual(state.count, 1)
	

	def test_create_button(self):
		button = self.counter.create_button("test", self.counter.increment)


		self.assertEqual(len(button.attributes), 2)
		self.assertEqual(len(button.children), 1)

		self.assertEqual(self.btnstyles.btn, button.attributes[1].value)
		self.assertEqual(self.counter.increment, button.attributes[0].value)
		self.assertEqual(button.children[0].name , "label")

	async def test_decrease(self):
		await self.counter.decrease()
		state = self.counter.get_state()
		self.assertEqual(state.count, -1)
	


	def test_node_structure(self):
		rendered_node = self.counter.render()
		# Check for top-level container and attributes 
		self.assertEqual(rendered_node.name, 'div')
		# Ensure h1 header is present
		header = rendered_node.children[0]
		self.assertEqual(header.name, 'h1')

		# Check controls div
		controls = rendered_node.children[1]
		self.assertEqual(controls.name, 'div')
		self.assertEqual(self.btnstyles.controls, rendered_node.children[1].attributes[0].value)

		# check parent node and nodes have unique node Id
		curr = [rendered_node]
		ids = []
		while curr:
			node = curr.pop()
			self.assertTrue(node.nodeId)
			self.assertNotIn(node.nodeId, ids)
			ids.append(node.nodeId)
			for i in node.children:
				if isinstance(i, Node):
					curr.append(i)


	def test_node_method_constructs_component_with_header_and_buttons(self):

		result = self.counter.render()

		# Assert
		self.assertEqual(result.name, "div")
		self.assertEqual(len(result.children), 2)
		self.assertEqual(result.children[0].name, "h1")
		self.assertEqual(len(result.children[0].children), 1)
		self.assertEqual(result.children[0].children[0].name, "text")
		self.assertEqual(len(result.children[0].children[0].children), 1)
		self.assertEqual(result.children[0].children[0].children[0].name, "data")
		self.assertEqual(len(result.children[1].children), 2)
		self.assertEqual(result.children[1].children[0].name, "button")
		self.assertEqual(result.children[1].children[0].children[0].name, "label")
		self.assertEqual(result.children[1].children[1].name, "button")
		self.assertEqual(result.children[1].children[1].children[0].name, "label")

	def test_create_button_with_empty_label(self):
		button = self.counter.create_button("", self.counter.increment)
		self.assertEqual(len(button.attributes), 2)
		self.assertEqual(len(button.children), 1)
		self.assertEqual(self.btnstyles.btn, button.attributes[1].value)
		self.assertEqual(self.counter.increment, button.attributes[0].value)
		self.assertEqual(button.children[0].name, "label")

	async def test_increment_multiple_times(self):
		# Increment the count multiple times
		for _ in range(5):
			await self.counter.increment()
		state = self.counter.get_state()
		self.assertEqual(state.count, 5)

	async def test_decrease_multiple_times(self):
		# Decrease the count multiple times
		for _ in range(3):
			await self.counter.decrease()
		state = self.counter.get_state()
		self.assertEqual(state.count, -3)


if __name__ == "__main__":
    unittest.main()


from zenaura.client.component import Component
from .lookup import VDomLookupTable

class HydratorVirtualDomAdapter(
    VDomLookupTable
):
    """
        hyderator adapter for all virtual dom operations
        methods should start with:
        hyd_vdom_
    """

    def hyd_vdom_update(self, comp: Component) -> None:
        """
            virtual dom operation : updates virtual dom of component
            args:
                comp: Component
        """
        self.zen_dom_table[comp.id] = comp.render()

    def hyd_vdom_delete(self, comp: Component) -> None:
        """
            virtual dom operation : deletes virtual dom of component
            args:
                comp: Component
        """
        del self.zen_dom_table[comp.id]

    def hyd_vdom_update_with_new_render(self, comp: Component, new_node):
        """
            update component with new node
            args:
                comp: Component
                new_node: Node
        """
        self.zen_dom_table[comp.id] = new_node




from .compiler_adapter import HydratorCompilerAdapter
from .real_dom_adapter import HydratorRealDomAdapter
from .virtual_dom_adapter import HydratorVirtualDomAdapter
from .tasker import HydratorTasker



class Hydrator(
    HydratorVirtualDomAdapter,
    HydratorCompilerAdapter,
    HydratorRealDomAdapter,
    HydratorTasker
):
    """
    Hydrator acts as the central communication hub between:

    1. **Virtual DOM and Compiler:**
        - Methods interacting with the compiler start with `hyd_comp_`, e.g., `hyd_comp_get_keyed_uuid`, `hyd_comp_compile_node`.
    2. **Virtual DOM:**
        - Methods interacting with Zenaura's virtual DOM start with `hyd_vdom_`, e.g., `hyd_vdom_update`, `hyd_vdom_delete`.
    3. **DOM:**
        - Methods interacting with the DOM start with `hyd_dom_`, e.g., `hyd_rdom_attach_to_root`.
    4. **Tasker:**
        - Tasks for updating the DOM are created in the updater and dequeued during the render lifecycle, enabling asynchronous DOM updates.

    Essentially, Hydrator bridges the gap between various components within the Zenaura framework, facilitating seamless communication and efficient DOM manipulation.
    """


import itertools
from .component import Component, UUIDManager
from typing import List


class Page:
    """
    Represents a page in a Zenaura application.

    A Page is a container for components that are displayed together. It manages the
    lifecycle of its child components and handles rendering them to the DOM.

    Attributes:
        count (int): A unique identifier for the page.
        id (str): A unique UUID for the page.
        children (List[Component]): The list of components that belong to this page.

    Raises:
        TypeError: If `children` is not a list or if any child is not a `Component`.
    """

    _page_count = itertools.count(0)

    def __init__(self, children: List[Component]):
        """
        Initializes a new Page instance.

        Args:
            children (List[Component]): The list of components to be added to the page.
        """

        self.count = next(self._page_count)
        self.id = UUIDManager.generate_uuid(self.__class__.__name__, self.count)
        self.children = children

        if not isinstance(self.children, list):
            raise TypeError("children must be a list")

        for child in self.children:
            if not isinstance(child, Component):
                raise TypeError("page children must be a Component")



from zenaura.client.tags import Attribute
from .sanitize import CompilerSanitizer
from typing import List 
import io 

sanitizer = CompilerSanitizer()

class AttributeProccessor(
    ):
    """
    This class is responsible for processing a list of `Attribute` objects and converting them into HTML-formatted attributes.

    Attributes:
        None
    """

    def process_attributes(
                self, 
                attrs: List[Attribute]
                ) -> str:
        """
        Processes a list of `Attribute` objects, converting them to HTML-formatted attributes.

        Args:
            attrs (List[Attribute]): A list of `Attribute` objects representing the attributes to be processed.

        Returns:
            str: A string containing the HTML-formatted attributes, ready to be included in a tag.

        Raises:
            TypeError: If the input `attrs` is not a list.
            ValueError: If any element in `attrs` is not an `Attribute` object.
        """

        if not isinstance(attrs, list):
            raise TypeError("`attrs` must be a list of Attribute objects.")

        for attr in attrs:
            if not isinstance(attr, Attribute):
                raise ValueError("Each element in `attrs` must be an Attribute object.")

        s = io.StringIO()  # Create a string buffer for building the output

        for i, attr in enumerate(attrs):
            attrKey = attr.key
            attrValue = attr.value
            if attrKey in self.attrKeyWords.keys():
                attrKey = self.attrKeyWords[attrKey]  # Apply keyword mapping

            # Add space only if it's not the first or last attribute
            if i == 0 or i == len(attrs) - 1:
                s.write(f' {attrKey}="{sanitizer.sanitize(attrValue)}"')
            else:
                s.write(f'{attrKey}="{sanitizer.sanitize(attrValue)}" ')

        res = s.getvalue()
        s.close()
        return res



from abc import ABC, abstractmethod

class Observer(ABC):
    """
    Abstract base class for observers that listen to changes in subjects.

    Observers are notified when a subject's state changes, allowing them to react accordingly.
    This pattern promotes loose coupling between subjects and observers, making the system more flexible and maintainable.

    **Key Features:**

    - **Loose Coupling:** Observers are not directly dependent on the subject's implementation, 
      allowing for easier modifications and additions.
    - **Flexibility:** Observers can be easily added or removed without affecting the subject.
    - **Maintainability:** The separation of concerns makes the code easier to understand and maintain.

    **Usage:**

    1. Define concrete observer classes that inherit from `Observer`.
    2. Implement the `update` method in each concrete observer to specify the desired behavior when notified.
    3. Attach the observers to the subject using the subject's `attach` method.
    4. When the subject's state changes, it will call the `update` method of all attached observers.

    **Example:**

    ```python
    class Subject:
        def __init__(self):
            self._observers = []

        def attach(self, observer):
            self._observers.append(observer)

        def notify(self, value):
            for observer in self._observers:
                observer.update(value)

    class ConcreteObserver(Observer):
        def update(self, value):
            print(f"Observer received value: {value}")

    # Create a subject and an observer
    subject = Subject()
    observer = ConcreteObserver()

    # Attach the observer to the subject
    subject.attach(observer)

    # Notify the observer with a value
    subject.notify("Hello, world!")
    ```

    **Additional Notes:**

    - The `update` method can receive any type of data, depending on the subject's implementation.
    - Observers can choose to ignore notifications if they are not interested in the updated value.
    - The subject can manage multiple observers and notify them individually or in groups.
    """

    @abstractmethod
    def update(self, value):
        """
        Update method to be implemented by concrete observers.

        This method is called by the subject when its state changes. The observer can use the provided value
        to perform any necessary actions, such as updating its own state or triggering other events.

        Parameters:
        value (dict): The updated value from the subject.

        Returns:
        None
        """
        pass


class Content:
    """
    Represents the content of an element.

    Attributes:
        firstChild (str): The text content of the first child element.
    """

    def __init__(self):
        self.firstChild = ""


class MockTextNode:
    """
    Represents a text node in the DOM.

    Attributes:
        nodeValue (str): The text content of the node.
        parentNode (MockElement): The parent element of the node.
    """

    def __init__(self, text):
        self.nodeValue = text
        self.parentNode = None


class MockElement:
    """
    Represents an HTML element in the DOM.

    Attributes:
        tagName (str): The tag name of the element.
        innerHTML (str): The HTML content of the element.
        outerHTML (str): The complete HTML representation of the element, including its children.
        attributes (dict): A dictionary of the element's attributes.
        childNodes (list): A list of the element's child nodes.
        parentNode (MockElement): The parent element of the node.
        content (Content): The content of the element.
    """

    def __init__(self, tag_name=None, innerHTML=""):
        self.tagName = tag_name
        self.innerHTML = innerHTML
        self.outerHTML = ""
        self.attributes = {}
        self.childNodes = []
        self.parentNode = None
        self.content = Content()

    def setAttribute(self, name, value):
        """
        Sets the value of an attribute on the element.

        Args:
            name (str): The name of the attribute.
            value (str): The value of the attribute.
        """

        self.attributes[name] = value

    def getAttribute(self, name):
        """
        Gets the value of an attribute on the element.

        Args:
            name (str): The name of the attribute.

        Returns:
            str: The value of the attribute, or None if the attribute does not exist.
        """

        return self.attributes.get(name)

    def removeAttribute(self, name):
        """
        Removes an attribute from the element.

        Args:
            name (str): The name of the attribute.
        """

        self.attributes.pop(name, None)

    def appendChild(self, child):
        """
        Appends a child node to the element.

        Args:
            child (MockElement or MockTextNode): The child node to append.
        """

        if child not in self.childNodes:
            self.childNodes.append(child)
            if isinstance(child, MockElement):
                child.parentNode = self

    def removeChild(self, child):
        """
        Removes a child node from the element.

        Args:
            child (MockElement or MockTextNode): The child node to remove.
        """

        if child in self.childNodes:
            self.childNodes.remove(child)
            child.parentNode = None

    def createElement(self, tag_name):
        """
        Creates a new MockElement with the given tag name.

        Args:
            tag_name (str): The tag name of the new element.

        Returns:
            MockElement: The newly created element.
        """

        return MockElement(tag_name)

    def __repr__(self):  # Helpful for debugging
        return f"<MockElement '{self.tagName}'>"


class MockDocument:  # Use MagicMock for flexibility
    """
    Represents the HTML document object model (DOM).

    Attributes:
        body (MockElement): The body element of the document.
        elementsById (dict): A dictionary of elements in the document, indexed by their ID.
        title (str): The title of the document.
    """

    def __init__(self):
        self.body = MockElement("body")  # Create a body element
        self.elementsById = {"root": self.body}  # Store elements by ID
        self.title = ""

    def getElementById(self, element_id):
        """
        Gets the element with the given ID.

        Args:
            element_id (str): The ID of the element.

        Returns:
            MockElement: The element with the given ID, or None if the element does not exist.
        """

        return self.elementsById.get(element_id)

    def createElement(self, tag_name):
        """
        Creates a new MockElement with the given tag name.

        Args:
            tag_name (str): The tag name of the new element.

        Returns:
            MockElement: The newly created element.
        """

        return MockElement(tag_name)

    def setElementById(self, element_id, element):
        """
        Sets the element with the given ID.

        Args:
            element_id (str): The ID of the element.
            element (MockElement): The element to set.
        """

        # for mocking purposes
        self.elementsById[element_id] = element

    def createTextNode(self, txt):
        """
        Creates a new MockTextNode with the given text.

        Args:
            txt (str): The text content of the new node.

        Returns:
            MockTextNode: The newly created node.
        """

        textNode = MockTextNode(txt)
        return textNode

    def querySelector(self, query: str):
        """
        Selects the first element that matches the given CSS selector.

        Args:
            query (str): The CSS selector to use.

        Returns:
            MockElement: The first element that matches the selector, or None if no element matches.
        """

        query = query.replace("[", "").replace("]", "").replace('"', "").split("=")
        id = query[-1]
        return self.elementsById[id]


class MockLocation:
    """
    Represents the location object of the browser window.

    Attributes:
        href (str): The URL of the current page.
        pathname (str): The path portion of the URL.
    """

    def __init__(self):
        self.href = "http://localhost:8000"  # Example
        self.pathname = ""


class MockWindowHistory:
    """
    Represents the history object of the browser window.

    Attributes:
        history (list): A list of the pages that have been visited.
    """

    def __init__(self):
        self.history = []

    def pushState(self, *args, **kwargs):
        """
        Adds a new entry to the history stack.

        Args:
            *args: Arguments to be passed to the history.pushState() method.
            **kwargs: Keyword arguments to be passed to the history.pushState() method.
        """

        self.history.append([args, kwargs])


class MockWindow:
    """
    Represents the browser window object.

    Attributes:
        innerWidth (int): The width of the window in pixels.
        innerHeight (int): The height of the window in pixels.
        location (MockLocation): The location object of the window.
        history (MockWindowHistory): The history object of the window.
    """

    def __init__(self):
        self.innerWidth = 1024
        self.innerHeight = 768
        self.location = MockLocation()
        self.history = MockWindowHistory()



from zenaura.client.tags import Node
from .attribute import AttributeProccessor
from .sanitize import CompilerSanitizer
import io
from zenaura.client.config import (
    ZENAURA_DOM_ATTRIBUTE,
    self_closing_tags
)


class Compiler(
    CompilerSanitizer,
    AttributeProccessor,
):
    """
    Compiles Zenui Nodes into their corresponding HTML representation.

    This class provides methods for:

    - Generating unique keyed UUIDs for each child within a component tree.
    - Compiling Zenui Nodes into HTML strings.
    - Sanitizing HTML content to prevent XSS vulnerabilities.
    - Processing attributes for HTML elements.

    Attributes:
        attrKeyWords (dict): A dictionary mapping attribute keywords to their corresponding HTML attribute names.
    """

    def __init__(self):
        """
        Initializes the Compiler instance with the default attribute keyword mapping.

        The default mapping includes:

        - `styles`: maps to `class` in HTML.

        You can customize this mapping by modifying the `attrKeyWords` attribute.
        """
        self.attrKeyWords = {
            "styles": "class",
        }

    def getKeyedUID(self, id, withAttribut=False, key=None):
        """
        Generates a unique keyed UUID for a child element within a component tree.

        This UUID is used to create a unique `ZENAURA_DOM_ATTRIBUTE` attribute for the element.

        Args:
            id (str): The unique ID of the parent component.
            withAttribut (bool, optional): Whether to include the `ZENAURA_DOM_ATTRIBUTE` attribute. Defaults to False.
            key (str or list, optional): The key or path of the child element. Defaults to None.

        Returns:
            str: The generated keyed UUID.

        Examples:
            ```python
            >>> compiler = Compiler()
            >>> compiler.getKeyedUID("my-component", key="child1")
            'my-componentchild1'
            >>> compiler.getKeyedUID("my-component", withAttribut=True, key=[0, 1])
            ' ZENAURA_DOM_ATTRIBUTE="my-component01"'
            ```
        """
        if isinstance(key, list):
            key = "".join(str(i) for i in key)  # Convert list to string

        if withAttribut:
            return f' {ZENAURA_DOM_ATTRIBUTE}="{id}{key}"'
        return f"{id}{key}"

    def compile(self, elm: Node, id=None, zenaura_dom_mode=False):
        """
        Compiles a Zenui Node into its corresponding HTML representation.

        Args:
            elm (Node): The Zenui Node object to compile.
            id (str, optional): The unique ID of the parent component. Used to generate keyed UUIDs.
            zenaura_dom_mode (bool, optional): Whether to add the `ZENAURA_DOM_ATTRIBUTE` attribute to the compiled HTML. Defaults to False.

        Returns:
            str: The compiled HTML string.

        Examples:
            ```python
            >>> compiler = Compiler()
            >>> node = Node("div", attributes=[Attribute("class", "my-class")], children=[Node("p", text="Hello, world!")])
            >>> compiler.compile(node)
            '<div class="my-class"><p>Hello, world!</p></div>'
            >>> compiler.compile(node, zenaura_dom_mode=True)
            '<div class="my-class" ZENAURA_DOM_ATTRIBUTE="div0"> <p ZENAURA_DOM_ATTRIBUTE="div00">Hello, world!</p></div>'
            ```
        """
        if elm.is_text_node or isinstance(elm, str):
            return self.sanitize(elm.text)

        tag = elm.name

        zenui_id = ""

        # Assign unique ID for Zenui DOM
        if isinstance(elm, Node) and zenaura_dom_mode:
            zenui_id = self.getKeyedUID(id, withAttribut=True, key=elm.path)

        # Get node attributes
        attributes = self.process_attributes(elm.attributes)

        if tag in self_closing_tags:
            return f"<{tag}{zenui_id}{attributes}>"

        # Start tag
        html = io.StringIO()
        html.write(f"<{tag}{zenui_id}{attributes}>")

        # Get children
        for child in elm.children:
            html.write(self.compile(child, id, zenaura_dom_mode))

        # Finish tag
        html.write(f"</{tag}>")

        return html.getvalue()


from abc import ABC, abstractmethod

class HTMLElement(ABC):
    """
    Abstract base class representing an HTML element.

    This class defines the common interface for all HTML elements,
    providing methods for manipulating child elements.

    Attributes:
        None

    Methods:
        appendChild(child: HTMLElement) -> None:
            Appends a child element to the current element.
        removeChild(child: HTMLElement) -> None:
            Removes a child element from the current element.
    """

    @abstractmethod
    def appendChild(self, child: "HTMLElement") -> None:
        """
        Appends a child element to the current element.

        Args:
            child: The child element to be appended.

        Raises:
            TypeError: If the provided child is not an instance of `HTMLElement`.
        """
        pass

    @abstractmethod
    def removeChild(self, child: "HTMLElement") -> None:
        """
        Removes a child element from the current element.

        Args:
            child: The child element to be removed.

        Raises:
            TypeError: If the provided child is not an instance of `HTMLElement`.
            ValueError: If the provided child is not a child of the current element.
        """
        pass



import sys
import unittest
from unittest.mock import patch  # For patching the global document object
from zenaura.client.tags import Node, Attribute  # Assuming these exist in your project
from zenaura.client.hydrator import HydratorRealDomAdapter
from zenaura.client.config import ZENAURA_DOM_ATTRIBUTE
from unittest.mock import patch, MagicMock
from tests.mocks.browser_mocks import MockDocument, MockElement, MockTextNode
from zenaura.client.tags import Node


sys.modules["pyscript"] = MagicMock()

class TestHydratorRealDomAdapter(unittest.TestCase):
    def setUp(self):
        self.mock_document = MockDocument()
        
        # Patch the global document object to use our mock
        self.patcher = patch("zenaura.client.hydrator.real_dom_adapter.document", self.mock_document)
        self.patcher.start()

        self.hydrator = HydratorRealDomAdapter()


    def tearDown(self):
        self.patcher.stop()  # Restore the original document after each test

    def test_hyd_rdom_create_element(self):
        virtual_node = Node("div")
        element = self.hydrator.hyd_rdom_create_element(virtual_node)
        self.assertTrue(element)
        self.assertEqual(element.tagName, "div")

    def test_hyd_rdom_attach_to_root(self):
        html = "<p>Test content</p>"
        self.hydrator.hyd_rdom_attach_to_root(html)
        root_element = self.mock_document.getElementById("root")
        self.assertEqual(root_element.innerHTML, html)

    def test_hyd_rdom_attach_to_mounted_comp(self):
        parent_div = self.mock_document.createElement("div")

        parent_div.setAttribute(ZENAURA_DOM_ATTRIBUTE, "parent-comp")
        self.mock_document.setElementById("parent_comp", parent_div)

        child_html = "<span>Child content</span>"
        self.hydrator.hyd_rdom_attach_to_mounted_comp("parent_comp", child_html)

        self.assertEqual(parent_div.outerHTML, child_html)
    
    def test_hyd_rdom_set_attribute(self):
        element = self.mock_document.createElement("div")
        element.setAttribute(ZENAURA_DOM_ATTRIBUTE, "my-comp")
        self.mock_document.setElementById("my-comp", element)

        attribute = Attribute(key="class", value="test-class")
        self.hydrator.hyd_rdom_set_attribute("my-comp", attribute)
        self.assertEqual(element.getAttribute("class"), "test-class")

    def test_hyd_rdom_remove_attribute(self):
        element = self.mock_document.createElement("div")
        element.setAttribute(ZENAURA_DOM_ATTRIBUTE, "my-comp")
        element.setAttribute("data-test", "some-value")
        self.mock_document.setElementById("my-comp", element)
        
        self.hydrator.hyd_rdom_remove_attribute("my-comp", "data-test")
        self.assertIsNone(element.getAttribute("data-test"))  # Ensure attribute is removed

    def test_hyd_rdom_remove_element(self):
        parent_div = self.mock_document.createElement("div")
        child_div = self.mock_document.createElement("div")
        child_div.setAttribute(ZENAURA_DOM_ATTRIBUTE, "child-comp")
        parent_div.appendChild(child_div)
        self.mock_document.setElementById("parent-comp", parent_div)
        self.mock_document.setElementById("child-comp", child_div)

        self.hydrator.hyd_rdom_remove_element("child-comp")
        self.assertEqual(len(parent_div.childNodes), 0)  # Child should be removed

    def test_hyd_rdom_append_child(self):
        parent_div = self.mock_document.createElement("div")
        parent_div.setAttribute(ZENAURA_DOM_ATTRIBUTE, "parent-comp")
        self.mock_document.setElementById("parent-comp", parent_div)

        child_html = "<span>Child content</span>"
        self.hydrator.hyd_rdom_append_child("parent-comp", child_html)
        self.assertEqual(len(parent_div.childNodes), 1)  # Should have one child

    def test_hyd_rdom_remove_child(self):
        parent_div = self.mock_document.createElement("div")
        parent_div.setAttribute(ZENAURA_DOM_ATTRIBUTE, "parent-comp")

        child_div = self.mock_document.createElement("div")
        child_div.setAttribute(ZENAURA_DOM_ATTRIBUTE, "child-comp")
        parent_div.appendChild(child_div)

        self.mock_document.setElementById("parent-comp", parent_div)
        self.mock_document.setElementById("child-comp", child_div)

        self.hydrator.hyd_rdom_remove_child("child-comp")
        self.assertEqual(child_div.outerHTML, "")  # Child should be removed
    
    #  New test for hyd_rdom_add_text_node
    def test_hyd_rdom_add_text_render(self):
        parent_div = self.mock_document.createElement("div")
        parent_div.setAttribute(ZENAURA_DOM_ATTRIBUTE, "parent-comp")
        self.mock_document.setElementById("parent-comp", parent_div)
        text_content = "Some text"
        self.hydrator.hyd_rdom_add_text_render("parent-comp", text_content)
        self.assertEqual(len(parent_div.childNodes), 1)
        self.assertEqual(parent_div.childNodes[0].nodeValue, text_content)


    def test_hyd_rdom_replace_inner_text(self):
        element = self.mock_document.createElement("div")
        element.setAttribute(ZENAURA_DOM_ATTRIBUTE, "my-comp")
        existing_text = "Existing text"
        self.mock_document.setElementById("my-comp", element)
        self.hydrator.hyd_rdom_replace_inner_text("my-comp", existing_text)  # Add initial text node

        new_text = "New text"
        self.hydrator.hyd_rdom_replace_inner_text("my-comp", new_text)

        # Assertions:
        self.assertEqual(element.textContent, new_text)  # Text should be replaced

if __name__ == "__main__":
    unittest.main()



import unittest
import asyncio
import sys
from unittest.mock import patch, MagicMock

from unittest.mock import patch, MagicMock

sys.modules["pyscript"] = MagicMock()

class TestDom(unittest.TestCase):

    @patch('pyscript.document')
    @patch('pyscript.window')
    def setUp(self, document, window):  # Run before each test
        from zenaura.client.hydrator.tasker import HydratorTasker
        self.tasker = HydratorTasker()

    def test_get_or_create_task_queue_new_queue(self):
        component_id = "comp-id1"
        queue = self.tasker.hyd_tsk_get_or_create_task_queue(component_id)
        self.assertIsInstance(queue, asyncio.Queue)
        self.assertEqual(self.tasker.queue_lookup[component_id], queue)

    def test_get_or_create_task_queue_existing_queue(self):
        component_id = "comp-id1"
        queue1 = self.tasker.hyd_tsk_get_or_create_task_queue(component_id)
        queue2 = self.tasker.hyd_tsk_get_or_create_task_queue(component_id)
        self.assertEqual(queue1, queue2)

    def test_enqueue_task_success(self):
        component_id = "comp-id3"
        queue = self.tasker.hyd_tsk_get_or_create_task_queue(component_id)
        task = MagicMock()
        result = self.tasker.hyd_tsk_enqueue_task(component_id, task)
        self.assertTrue(result)
        self.assertEqual(queue.qsize(), 1)
        dequeued_task = self.tasker.hyd_tsk_dequeue_task(component_id)
        self.assertEqual(dequeued_task, task)
        self.assertEqual(queue.qsize(),0)
        self.tasker.hyd_tsk_dequeue_task(component_id)
        # task queue cleaned after emptied
        self.assertEqual(self.tasker.queue_lookup[component_id], None)
    
    def test_dequeue_task_empty_queue(self):
        component_id = "comp-id5"
        queue = self.tasker.hyd_tsk_get_or_create_task_queue(component_id)

        result = self.tasker.hyd_tsk_dequeue_task(component_id)

        self.assertTrue(callable(result))
        self.assertEqual(self.tasker.queue_lookup[component_id], None)
    
    def test_dequeue_task_non_existent_queue(self):
        component_id = "comp-id6"
        queue = self.tasker.hyd_tsk_get_or_create_task_queue(component_id)

        result = self.tasker.hyd_tsk_dequeue_task(component_id)
        self.assertTrue(callable(result))
        self.assertEqual(self.tasker.queue_lookup[component_id], None)

    def test_dequeue_task_multiple_tasks(self):
        component_id = "comp-id7"
        queue = self.tasker.hyd_tsk_get_or_create_task_queue(component_id)
        task1 = MagicMock()
        task2 = MagicMock()
        task3 = MagicMock()
        queue.put_nowait(task1)
        queue.put_nowait(task2)
        queue.put_nowait(task3)
        dequeued_task1 = self.tasker.hyd_tsk_dequeue_task(component_id)
        dequeued_task2 = self.tasker.hyd_tsk_dequeue_task(component_id)
        dequeued_task3 = self.tasker.hyd_tsk_dequeue_task(component_id)
        self.tasker.hyd_tsk_dequeue_task(component_id)
        self.assertEqual(dequeued_task1, task1)
        self.assertEqual(dequeued_task2, task2)
        self.assertEqual(dequeued_task3, task3)
        self.assertEqual(queue.qsize(), 0)
        self.assertEqual(self.tasker.queue_lookup[component_id], None)





from .mount import Mount
from .render import Render

class Dom(
    Mount,
    Render
):
    """
    Zenaura virtual DOM.

    This class combines the functionalities of `Mount` and `Render` classes to provide a comprehensive virtual DOM implementation.

    **Attributes:**

    * **Mount:** Inherits all attributes and methods from the `Mount` class, which handles the mounting lifecycle of components.
    * **Render:** Inherits all attributes and methods from the `Render` class, which handles the rendering lifecycle of components.

    **Methods:**

    * **mount(comp, container):** Mounts the component to the specified container element in the real DOM.
    * **render(comp):** Renders the component to a virtual DOM tree.
    * **update(prev_comp, new_comp):** Updates the real DOM based on the differences between the previous and new virtual DOM trees.

    **Usage:**

    ```python
    # Create a Dom instance
    dom = Dom()

    # Create a component
    comp = MyComponent()

    # Mount the component to a container element
    dom.mount(comp, container)

    # Render the component
    dom.render(comp)

    # Update the component
    dom.update(prev_comp, new_comp)
    ```
    """



from .lifecycles.mount import MountLifeCycles
from .error import GracefulDegenerationLifeCycleWrapper
from zenaura.client.page import Page
from zenaura.client.hydrator import HydratorRealDomAdapter
import traceback

rdom_hyd = HydratorRealDomAdapter()

class Mount(
    GracefulDegenerationLifeCycleWrapper,
    MountLifeCycles,
    ):
    """
    Mounts a single `Page` instance to the DOM.

    This class handles the mounting of a single `Page` instance to the DOM. It ensures that only one page is mounted at a time and provides lifecycle methods for handling the mounting process.

    **Lifecycle:**

    1. **Server-side:**
        - The server hydrates all app pages and overwrites the `index.html` file.
        - The `App` class toggles the visibility of the mounted page.
    2. **Client-side:**
        - The `Mount` class triggers the `attached` lifecycle method for the page components.

    **Error Handling:**

    - If an error occurs during mounting, the `componentDidCatchError` method is called with the error message.
    - This method allows components to handle errors gracefully by returning a new component to display in place of the original component.
    - If the component does not have a `componentDidCatchError` method, a default error message component is displayed.

    **Parameters:**

    - `page`: An instance of the `Page` class.

    **Returns:**

    None
    """

    async def mount(self, page: Page) -> None:
        """
        Mounts the given `Page` instance to the DOM.

        This method attempts to mount the provided `Page` instance to the DOM. It iterates through the page's children, updating their state in the virtual DOM and triggering the `attached` lifecycle method for each component.

        **Error Handling:**

        - If an error occurs during mounting, the `componentDidCatchError` method is called with the error message.
        - This method allows components to handle errors gracefully by returning a new component to display in place of the original component.
        - If the component does not have a `componentDidCatchError` method, a default error message component is displayed.

        **Parameters:**

        - `page`: An instance of the `Page` class.

        **Returns:**

        None
        """

        try:
            for comp in page.children:
                # Update state in vdom
                self.hyd_vdom_update(comp)
                # Trigger attached for page components
                await self.attached(comp)

        except Exception as e:
            self.componentDidCatchError(page.children[0], traceback.format_exc())



import sqlite3

class PersistenceRegistry:
    """
    A class for managing persistence of UUIDs and HTML content in a SQLite database.

    This class provides methods for:

    * Creating and managing tables in the database.
    * Mapping UUIDs to integer IDs and vice versa.
    * Storing and retrieving HTML content associated with components.

    Attributes:
        conn (sqlite3.Connection): The connection to the SQLite database.
    """

    def __init__(self, db_file):
        """
        Initializes a new PersistenceRegistry instance.

        Args:
            db_file (str): The path to the SQLite database file.
        """

        self.conn = sqlite3.connect(db_file)
        self.create_tables()

    def create_tables(self):
        """
        Creates the necessary tables in the database if they don't already exist.

        The following tables are created:

        * `uuid_mapping`: Maps UUIDs to integer IDs.
        * `html_mapping`: Maps component IDs to HTML content.
        """

        cursor = self.conn.cursor()

        cursor.execute('''CREATE TABLE IF NOT EXISTS uuid_mapping (
                            uuid TEXT PRIMARY KEY,
                            integer_id INTEGER UNIQUE
                        )''')

        cursor.execute('''CREATE TABLE IF NOT EXISTS html_mapping (
                            component_id TEXT PRIMARY KEY,
                            html_content TEXT
                        )''')

        self.conn.commit()

    def find_or_create_uuid_integer_mapping(self, uuid_str, integer_id):
        """
        Finds the existing UUID for the given integer ID, or creates a new mapping if none exists.

        Args:
            uuid_str (str): The UUID string.
            integer_id (int): The integer ID.

        Returns:
            str: The existing or newly created UUID.
        """

        existing_uuid = self.retrieve_uuid(integer_id)
        if existing_uuid:
            return existing_uuid
        return self.insert_uuid(uuid_str, integer_id)

    def insert_uuid(self, uuid_str, integer_id):
        """
        Inserts a new mapping between a UUID and an integer ID into the database.

        Args:
            uuid_str (str): The UUID string.
            integer_id (int): The integer ID.
        """

        cursor = self.conn.cursor()
        cursor.execute('''INSERT INTO uuid_mapping (uuid, integer_id)
                          VALUES (?, ?)''', (uuid_str, integer_id))
        self.conn.commit()

    def retrieve_uuid(self, integer_id):
        """
        Retrieves the UUID associated with the given integer ID.

        Args:
            integer_id (int): The integer ID.

        Returns:
            str: The UUID string, or None if no mapping exists.
        """

        cursor = self.conn.cursor()
        cursor.execute('''SELECT uuid FROM uuid_mapping
                          WHERE integer_id = ?''', (integer_id,))
        row = cursor.fetchone()
        return row[0] if row else None

    def insert_html(self, component_id, html_content):
        """
        Inserts the HTML content associated with a component into the database.

        Args:
            component_id (str): The component ID.
            html_content (str): The HTML content.
        """

        cursor = self.conn.cursor()
        cursor.execute('''INSERT INTO html_mapping (component_id, html_content)
                              VALUES (?, ?)''', (component_id, html_content))
        self.conn.commit()

    def retrive_html(self, component_id):
        """
        Retrieves the HTML content associated with a component from the database.

        Args:
            component_id (str): The component ID.

        Returns:
            str: The HTML content, or None if no content is found.
        """

        cursor = self.conn.cursor()
        cursor.execute('''SELECT uuid FROM html_mapping
                          WHERE component_id = ?''', (component_id,))
        row = cursor.fetchall()
        print("row", row)
        return row[0] if row else None

registry = PersistenceRegistry("./persistance.db")



from collections import defaultdict
import asyncio
class HydratorTasker:
    """
    Manages tasks generated by the Zenaura diffing algorithm for updating both the real DOM and the virtual DOM.

    This class provides methods for:

    - Creating and retrieving task queues for specific component IDs.
    - Enqueuing tasks for components.
    - Dequeuing tasks for components.

    **Key Features:**

    - Efficient task management for real and virtual DOM updates.
    - Thread-safe task queue operations.
    - Automatic cleanup of empty task queues.

    **Usage:**

    1. Use `hyd_tsk_get_or_create_task_queue` to get or create a task queue for a specific component ID.
    2. Use `hyd_tsk_enqueue_task` to enqueue a task for a component.
    3. Use `hyd_tsk_dequeue_task` to dequeue a task for a component.

    **Example:**

    ```python
    tasker = HydratorTasker()

    # Get or create a task queue for component ID "my-component"
    task_queue = tasker.hyd_tsk_get_or_create_task_queue("my-component")

    # Enqueue a task for the component
    task_queue.put_nowait(my_task)

    # Dequeue a task for the component
    task = task_queue.get_nowait()

    # Execute the task
    task()
    ```
    """

    queue_lookup = defaultdict(lambda: None)

    def __init__(self):
        """
        Initializes the HydratorTasker instance.
        """
        pass

    def hyd_tsk_get_or_create_task_queue(self, component_id) -> asyncio.Queue:
        """
        Gets or creates a task queue for the specified component ID.

        Args:
            component_id (str): The ID of the component.

        Returns:
            asyncio.Queue: The task queue for the component.
        """
        try:
            if component_id not in self.queue_lookup:
                self.queue_lookup[component_id] = asyncio.Queue()
                return self.queue_lookup[component_id]
            else:
                return self.queue_lookup[component_id]
        except KeyError:
            return False

    def hyd_tsk_enqueue_task(self, component_id, task):
        """
        Enqueues a task for the specified component ID.

        Args:
            component_id (str): The ID of the component.
            task (callable): The task to be enqueued.

        Returns:
            bool: True if the task was successfully enqueued, False otherwise.
        """
        comp_queue = self.queue_lookup[component_id]
        try:
            comp_queue.put_nowait(task)
            return True
        except asyncio.QueueFull:
            return False

    async def hyd_tsk_do_nothing(self):
        """
        A placeholder function used when a task queue is empty.
        """
        pass

    def hyd_tsk_dequeue_task(self, component_id):
        """
        Dequeues a task for the specified component ID.

        Args:
            component_id (str): The ID of the component.

        Returns:
            callable: The dequeued task, or `hyd_tsk_do_nothing` if the queue is empty.
        """
        comp_queue = self.queue_lookup[component_id]
        try:
            task = comp_queue.get_nowait()
            return task
        except asyncio.QueueEmpty:
            # Clean up and return the placeholder function
            del self.queue_lookup[component_id]
            return self.hyd_tsk_do_nothing



import asyncio 
from zenaura.client.tags import Node, HTMLElement, Attribute
from zenaura.client.page import Page 
from zenaura.client.component import Component
from zenaura.client.config import ZENAURA_DOM_ATTRIBUTE
from zenaura.client.mocks import MockDocument, MockWindow
try :
    from pyscript import document, window
    from pyodide.ffi import create_proxy
    in_browser = True

except ImportError:
    document = MockDocument() 
    window = MockWindow() 
    create_proxy = lambda x : x 
    in_browser = False
from typing import Dict

class HydratorRealDomAdapter:
    # for testing 
    """
        hyderator adapter for all real dom operations
        methods should start with:
        hyd_romp_
    """


    def hyd_rdom_create_element(self, virtual_node: Node) -> HTMLElement:
        """
        DOM operation: creates html element and returns it as HTMLElement

        Args:
            virtual_node: Node - the virtual node representing the element to be created

        Returns:
            HTMLElement - the created HTML element
        """
        element = document.createElement(virtual_node.name)
        
        return element
    
    def hyd_rdom_attach_to_root(self, html : str) -> None:
        """
        DOM operation: attaches compiled_comp to mounted_comp_id

        Args:
            html: str - the HTML string to be attached to the root element
        """
        document.getElementById("root").innerHTML = html

    def hyd_rdom_attach_to_mounted_comp(
            self,
            mounted_comp_id: str,
            html: str
    ):
        """
        DOM operation: attaches compiled_comp to mounted_comp_id

        Args:
            mounted_comp_id: str - the ID of the previously mounted component
            html: str - the HTML string to be attached to the mounted component
        """
        foundNode = document.querySelector(f'[{ZENAURA_DOM_ATTRIBUTE}="{mounted_comp_id}"]')
        if foundNode:
            foundNode.outerHTML = html

    def hyd_rdom_set_attribute(self, mounted_comp_id: str, attribute: Attribute) -> None:
        """
        DOM operation: Sets attributes on an HTML element.

        Args:
            mounted_comp_id: The ID of the element to modify.
            attributes: A dictionary of attribute names and their values.
        """
        element = document.querySelector(f'[{ZENAURA_DOM_ATTRIBUTE}="{mounted_comp_id}"]')
        if element:
            element.setAttribute(attribute.key, attribute.value)


    def hyd_rdom_remove_attribute(self, mounted_comp_id: str, attribute_name: str) -> None:
        """
        DOM operation: Removes an attribute from an HTML element.

        Args:
            mounted_comp_id: The ID of the element.
            attribute_name: The name of the attribute to remove.
        """
        element = document.querySelector(f'[{ZENAURA_DOM_ATTRIBUTE}="{mounted_comp_id}"]')
        if element:
            element.removeAttribute(attribute_name)

    def hyd_rdom_remove_element(self, mounted_comp_id: str) -> None:
        """
        DOM operation: removes an element from the DOM

        Args:
            mounted_comp_id: str - the ID of the element to be removed
        """
        element = document.querySelector(f'[{ZENAURA_DOM_ATTRIBUTE}="{mounted_comp_id}"]')
        if element:
            element.parentNode.removeChild(element)
   
    
    def hyd_rdom_append_child(self, mounted_comp_id:str, child_html:str) -> None:
        """
        DOM operation: appends a child to an element

        Args:
            mounted_comp_id: str - the ID of the parent element
            child_html: str - the HTML string of the child element to be appended
        """
        element = document.querySelector(f'[{ZENAURA_DOM_ATTRIBUTE}="{mounted_comp_id}"]')
        if element:
            child_node = document.createElement("template")
            child_node.innerHTML = child_html
            element.appendChild(child_node.content.firstChild)

    def hyd_rdom_append_child_after(self, parent_node_id, child_node_id, child_html) -> None:
        """
        DOM operation: appends a child to an element after a specific child

        Args:
            parent_node_id: str - the ID of the parent element
            child_node_id: str - the ID of the child element to insert after
            child_html: str - the HTML string of the child element to be appended
        """
        element = document.querySelector(f'[{ZENAURA_DOM_ATTRIBUTE}="{parent_node_id}"]')
        if element : # if parent exists
            child_node = document.createElement("template")
            child_node.innerHTML = child_html
            curr_node = child_node.content.firstChild
            child_index = int( child_node_id[-1])
            prev_child = child_index - 1 # child on the dom to insert the new child after it
            prev_child_id = child_node_id[:-1] + str(prev_child) # parent-childIndex
            prev_child = document.querySelector(f'[{ZENAURA_DOM_ATTRIBUTE}="{prev_child_id}"]')
            if prev_child: # insert after the current child:
                element.insertBefore(curr_node, prev_child.nextSibling)
            else: # parent is a leaf no children
                element.append(curr_node)
            pass

    def hyd_rdom_remove_child(self, child_id:str) -> None:
        """
        DOM operation: removes a child of an element

        Args:
            child_id: str - the ID of the child element to be removed
        """
        child_node = document.querySelector(f'[{ZENAURA_DOM_ATTRIBUTE}="{child_id}"]')
        if child_node:
            child_node.outerHTML = ""


    def hyd_rdom_add_text_render(self, mounted_comp_id: str, text_content: str) -> None:
        """
        DOM operation: Adds a text node to an element.

        Args:
            mounted_comp_id: str - the ID of the element to add the text node to
            text_content: str - the text content of the text node
        """
        element = document.querySelector(f'[{ZENAURA_DOM_ATTRIBUTE}="{mounted_comp_id}"]')
        if element:
            text_node = document.createTextNode(text_content)
            element.appendChild(text_node)
            

    def hyd_rdom_replace_inner_text(self, mounted_comp_id: str, new_text: str) -> None:
        """
        DOM operation: Replaces the inner text content of an element.

        Args:
            mounted_comp_id: str - the ID of the element to replace the text content of
            new_text: str - the new text content
        """
        element = document.querySelector(f'[{ZENAURA_DOM_ATTRIBUTE}="{mounted_comp_id}"]')
        if element:
            element.textContent = new_text

    def hyd_rdom_is_interactive(self) -> bool:
        """
        DOM operation: checks if real dom  is interactive

        Returns:
            bool - True if the DOM is interactive, False otherwise
        """
        return document.readyState == "interactive"

    def hyd_rdom_is_complete(self) -> bool:
        """
        DOM operation: checks if real dom  is complete

        Returns:
            bool - True if the DOM is complete, False otherwise
        """
        return document.readyState == "complete"
    
    def hyd_rdom_is_loading(self) -> bool:
        """
        DOM operation: checks if real dom  is complete

        Returns:
            bool - True if the DOM is complete, False otherwise
        """
        return document.readyState == "load"
    
    def hyd_rdom_is_content_loaded(self) -> bool:
        """
        DOM operation: checks if real dom  is content loaded

        Returns:
            bool - True if the DOM is content loaded, False otherwise
        """
        return document.readyState == "DOMContentLoaded"

    async def hyd_rdom_wait_for_dom_content_loaded(self):
        """
        Waits for the DOM to be content loaded.
        """
        if not in_browser:
            return 
        while True:
            await asyncio.sleep(0.001)
            if document.readyState=="complete":
                break
        

    def hyd_rdom_toggle_pages_visibilty(self, previous_page : Page, current_page : Page ):
        """
        Toggles the visibility of the previous and current pages.

        Args:
            previous_page: Page - the previously mounted page
            current_page: Page - the currently mounted page
        """
        p_page = document.querySelector(f'[data-zenaura="{previous_page.id}"]')
        if p_page:
            p_page.hidden = True
        curr_page = document.querySelector(f'[data-zenaura="{current_page.id}"]')
        if curr_page:
            curr_page.hidden = False # Update the title


from zenaura.client.hydrator import Hydrator
from .operations import *
from typing import List
from zenaura.client.tags.attribute import Attribute

class Updater(
    Hydrator
    ):
    """
    This class is responsible for updating the real DOM based on the differences identified by the Searcher.

    It receives a list of patches from the Searcher and creates corresponding tasks for each patch. These tasks are then enqueued for execution by the Hydrator's tasker.

    Attributes:
        None
    """

    async def update(self, patches: List, id: str) -> None:
        """
        Updates the real DOM based on the provided patches.

        Args:
            patches: A list of patches containing the operation name, the new child node, the path of the child, and the context for the updater.
            id: The ID of the component to be updated.
        """

        while patches:
            prev_node_id, diffed_node, path, op = patches.pop(0)
            # print("operation", op, prev_node_id)

            if op["name"] == ADD_NODE:
                async def task(dn=diffed_node, ci=id):
                    """
                    Adds a new node to the real DOM.

                    Args:
                        dn: The new child node to be added.
                        ci: The ID of the component to which the node should be added.
                    """
                    compiled_html = self.hyd_comp_compile_children(dn, ci, True)
                    parent_id = id + dn.path[0:-2]  # last two digit in keyed uid is always level, index, so parent lives at the before
                    child_id = id + dn.path
                    self.hyd_rdom_append_child_after(parent_id, child_id, compiled_html)
                self.hyd_tsk_enqueue_task(id, task)

            if op["name"] == REMOVE_NODE:
                async def task(d=diffed_node, c=prev_node_id):
                    """
                    Removes a node from the real DOM.

                    Args:
                        d: The node to be removed.
                        c: The ID of the component containing the node.
                    """
                    compiled_html = self.hyd_comp_compile_children(d, c, True)
                    # print("ADD_NODE compiled html", compiled_html)
                    self.hyd_rdom_remove_child(prev_node_id)
                self.hyd_tsk_enqueue_task(id, task)

            if op["name"] == NODE_INNER_TEXT:
                # avoid late binding
                async def task(id=prev_node_id, c=op["context"]["text"]):
                    """
                    Updates the inner text of a node in the real DOM.

                    Args:
                        id: The ID of the node to be updated.
                        c: The new inner text.
                    """
                    self.hyd_rdom_replace_inner_text(id, c)
                self.hyd_tsk_enqueue_task(id, task)

            if op["name"] == ADD_ATTRIBUTE:
                async def task(id=prev_node_id, op1=Attribute(op["context"]["attr_name"], op["context"]["attr_value"])):
                    """
                    Adds an attribute to a node in the real DOM.

                    Args:
                        id: The ID of the node to be updated.
                        op1: The attribute to be added.
                    """
                    self.hyd_rdom_set_attribute(id, op1)
                self.hyd_tsk_enqueue_task(id, task)

            if op["name"] == REMOVE_ATTRIBUTE:
                async def task(id=prev_node_id, op1=op["context"]["attr_name"]):
                    """
                    Removes an attribute from a node in the real DOM.

                    Args:
                        id: The ID of the node to be updated.
                        op1: The name of the attribute to be removed.
                    """
                    self.hyd_rdom_remove_attribute(id, op1)
                self.hyd_tsk_enqueue_task(id, task)

            if op["name"] == REPLACE_ATTRIBUTE:
                async def task(id=prev_node_id, op1=op["context"]["attr_name"]):
                    """
                    Replaces an attribute on a node in the real DOM.

                    Args:
                        id: The ID of the node to be updated.
                        op1: The name of the attribute to be replaced.
                    """
                    self.hyd_rdom_remove_attribute(id, op1)
                self.hyd_tsk_enqueue_task(id, task)



from zenaura.client.component import Component, Reuseable
from zenaura.client.tags import Node
from zenaura.client.compiler import compiler
from zenaura.client.hydrator import Hydrator

@Reuseable
class DefaultDomErrorComponent(Component):
    """
    Displays a default error message component.

    Attributes:
        error_message (str): The error message to display.

    Methods:
        render(): Returns a Node representing the error message.
    """

    def __init__(self, error_message):
        super().__init__()
        self.error_message = error_message

    def render(self):
        return Node("div", children=[Node(text=str(self.error_message))])


class GracefulDegenerationLifeCycleWrapper(
    Hydrator
):
    """
    Wraps components to handle errors gracefully.

    This class provides a `componentDidCatchError` method that allows components
    to handle errors gracefully. If a component throws an error, the
    `componentDidCatchError` method will be called with the error message. The
    component can then return a new component to display in place of the original
    component.

    If the component does not have a `componentDidCatchError` method, a default
    error message component will be displayed.
    """

    def componentDidCatchError(self, comp, error) -> None:
        """
        Handles errors gracefully.

        This method is called when a component throws an error. It allows the
        component to handle the error gracefully by returning a new component to
        display in place of the original component.

        Args:
            comp (Component): The component that threw the error.
            error (Exception): The error that was thrown.
        """

        # Cleanup the Zen DOM table.
        self.zen_dom_table.clear()

        if hasattr(comp, "componentDidCatchError"):
            # Call the component's `componentDidCatchError` method.
            error_comp = comp.componentDidCatchError(str(error))

            # Compile and render the error component.
            compiled_comp = self.hyd_comp_compile_render(error_comp)

            # Attach the compiled component to the real DOM.
            self.hyd_rdom_attach_to_root(compiled_comp)

            # Update the virtual DOM with the new render.
            self.hyd_vdom_update_with_new_render(comp, error_comp.render())

        else:
            # Create a default error message component.
            error_comp = DefaultDomErrorComponent(error_message=str(error))

            # Compile and render the default error component.
            compiled_comp = self.hyd_comp_compile_render(error_comp)

            # Attach the compiled component to the real DOM.
            self.hyd_rdom_attach_to_root(compiled_comp)

            # Update the virtual DOM with the new render.
            self.hyd_vdom_update_with_new_render(comp, error_comp.render())



import unittest
from unittest.mock import MagicMock 
from zenaura.client.observer import Observer, Subject

observer1 = MagicMock(spec=Observer)

observer2 = MagicMock(spec=Observer)

observer1.update = MagicMock(return_value=3)

observer2.update = MagicMock(return_value=3)

class TestSubject(unittest.TestCase):
    def test_state_set_to_none(self):
        subject = Subject()
        subject.state = None

    def test_state_set_to_empty_dict(self):
        subject = Subject()
        subject.state = {}

    def test_notify_observers(self):
        subject = Subject()

        subject.attach(observer1)
        subject.attach(observer2)
        subject.state = None
        observer1.update.assert_called_once_with(None)
        observer2.update.assert_called_once_with(None)

    def test_state_set_to_none(self):
        subject = Subject()
        subject.state = None
        self.assertIsNone(subject.state)

    def test_state_set_to_empty_dict(self):
        subject = Subject()
        subject.state = {}
        self.assertDictEqual(subject.state, {})

    def test_detach_observer(self):
        subject = Subject()
        observer = MagicMock()
        subject.attach(observer)
        self.assertIn(observer, subject._observers)
        subject.detach(observer)
        self.assertNotIn(observer, subject._observers)


if __name__ == '__main__':
    unittest.main()


from collections import defaultdict

class VDomLookupTable:
    """
        VDomLookupTable:

        This class manages two lookup tables:

        1. **zen_dom_table**:
            - Optimizes general tree structure by storing each component's unique ID and its children with the `ZENAURA_DOM_ATTRIBUTE`.
            - Provides efficient operations:
                - Searching for mounted component's previous state: O(1)
                - Deleting unmounted components: O(1)
                - Inserting mounted components: O(1)
            - Applies the same time complexity to tree structure operations:
                - Search: O(n)
                - Insert: O(n)
                - Delete: O(n)
                - Update: O(n)
            - Enables efficient memory management by deleting all components when a page is unmounted.

        2. **zenaura_prev_page_instance**:
            - Stores the previously mounted page instance.
            - Leverages Zenaura's single-page mounting (per route path) for efficient memory management.
            - When a new page is mounted:
                - Retrieves the previous page instance.
                - Iterates over its components and deletes them from `zen_dom_table`.

    """
    zen_dom_table = defaultdict(str)
    prev_page_instance = None
    zen_pre_compiled = defaultdict(str)
